"""
Core recipes for the NewtonNet code
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ase.optimize import FIRE
from ase.vibrations.data import VibrationsData
from monty.dev import requires

from quacc import SETTINGS, job
from quacc.schemas import fetch_atoms
from quacc.schemas.ase import (
    summarize_opt_run,
    summarize_run,
    summarize_thermo,
    summarize_vib_run,
)
from quacc.utils.calc import run_ase_opt, run_calc
from quacc.utils.dicts import merge_dicts
from quacc.utils.thermo import ideal_gas

try:
    from sella import Sella
except ImportError:
    Sella = None

try:
    from newtonnet.utils.ase_interface import MLAseCalculator as NewtonNet
except ImportError:
    NewtonNet = None

if TYPE_CHECKING:
    from ase import Atoms

    from quacc.schemas.ase import OptSchema, RunSchema, ThermoSchema, VibSchema

    class FreqSchema(VibSchema):
        thermo: ThermoSchema


@job
@requires(NewtonNet, "NewtonNet must be installed. Try pip install quacc[newtonnet]")
def static_job(
    atoms: Atoms | dict,
    calc_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> RunSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    calc_swaps
        Dictionary of custom kwargs for the newtonnet calculator. Overrides the
        following defaults:

        ```python
        {
            "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
            "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
        }
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    RunSchema
        Dictionary of results, specified in `quacc.schemas.ase.RunSchema`
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}

    defaults = {
        "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
        "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
    }
    flags = merge_dicts(defaults, calc_swaps)

    atoms.calc = NewtonNet(**flags)
    final_atoms = run_calc(atoms, copy_files=copy_files)

    return summarize_run(
        final_atoms,
        input_atoms=atoms,
        additional_fields={"name": "NewtonNet Static"},
    )


@job
@requires(NewtonNet, "NewtonNet must be installed. Try pip install quacc[newtonnet]")
def relax_job(
    atoms: Atoms | dict,
    calc_swaps: dict | None = None,
    opt_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> OptSchema:
    """
    Relax a structure.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    calc_swaps
        Dictionary of custom kwargs for the newtonnet calculator. Overrides the
        following defaults:

        ```python
        {
            "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
            "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
        }
        ```
    opt_swaps
        Optional swaps for the optimization parameters. Overrides the following
        defaults:

        ```python
        {"fmax": 0.01, "max_steps": 1000, "optimizer": Sella or FIRE}
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    OptSchema
        Dictionary of results, specified in `quacc.schemas.ase.OptSchema``
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}
    opt_swaps = opt_swaps or {}

    defaults = {
        "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
        "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
    }
    opt_defaults = {"fmax": 0.01, "max_steps": 1000, "optimizer": Sella or FIRE}

    flags = merge_dicts(defaults, calc_swaps)
    opt_flags = merge_dicts(opt_defaults, opt_swaps)

    if "sella.optimize" in opt_flags.get("optimizer", FIRE).__module__:
        opt_flags["order"] = 0

    atoms.calc = NewtonNet(**flags)
    dyn = run_ase_opt(atoms, copy_files=copy_files, **opt_swaps)

    return _add_stdev_and_hess(
        summarize_opt_run(dyn, additional_fields={"name": "NewtonNet Relax"})
    )


@job
@requires(NewtonNet, "NewtonNet must be installed. Try pip install quacc[newtonnet]")
def freq_job(
    atoms: Atoms | dict,
    temperature: float = 298.15,
    pressure: float = 1.0,
    calc_swaps: dict | None = None,
) -> FreqSchema:
    """
    Perform a frequency calculation using the given atoms object.

    Parameters
    ----------
    atoms
        The atoms object representing the system.
    temperature
        The temperature for the thermodynamic analysis.
    pressure
        The pressure for the thermodynamic analysis.
    calc_swaps
        Optional swaps for the calculator. Overrides the following
        defaults:

        ```python
        {
            "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
            "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
        }
        ```

    Returns
    -------
    FreqSchema
        Dictionary of results
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}

    defaults = {
        "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
        "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
    }
    flags = merge_dicts(defaults, calc_swaps)

    ml_calculator = NewtonNet(**flags)
    atoms.calc = ml_calculator

    ml_calculator.calculate(atoms)
    hessian = ml_calculator.results["hessian"]
    vib = VibrationsData(atoms, hessian)
    vib_summary = summarize_vib_run(
        vib, additional_fields={"name": "NewtonNet Frequency Analysis"}
    )

    igt = ideal_gas(
        atoms, vib.get_frequencies(), energy=ml_calculator.results["energy"]
    )
    vib_summary["thermo"] = summarize_thermo(
        igt, temperature=temperature, pressure=pressure
    )

    return vib_summary


def _add_stdev_and_hess(summary: dict) -> dict:
    """
    Calculate and add standard deviation values and Hessians to the summary.

    This function takes a summary dictionary containing information about a
    molecular trajectory and calculates the standard deviation of various
    properties using the NewtonNet machine learning calculator. It adds the
    calculated standard deviation values and Hessians to each configuration in
    the trajectory.

    Parameters
    ----------
    summary
        A dictionary containing information about the molecular trajectory.

    Returns
    -------
    Dict
        The modified summary dictionary with added standard deviation and
        Hessian values.
    """

    for conf in summary["trajectory"]:
        ml_calculator = NewtonNet(
            model_path=SETTINGS.NEWTONNET_MODEL_PATH,
            settings_path=SETTINGS.NEWTONNET_CONFIG_PATH,
        )
        ml_calculator.calculate(conf["atoms"])
        conf["hessian"] = ml_calculator.results["hessian"]
        conf["energy_std"] = ml_calculator.results["energy_disagreement"]
        conf["forces_std"] = ml_calculator.results["forces_disagreement"]
        conf["hessian_std"] = ml_calculator.results["hessian_disagreement"]

    return summary



"""Schemas for storing metadata about Atoms objects"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from ase.atoms import Atom, Atoms
from emmet.core.structure import MoleculeMetadata, StructureMetadata
from monty.json import jsanitize
from pymatgen.io.ase import AseAtomsAdaptor

from quacc.utils.atoms import copy_atoms
from quacc.utils.dicts import clean_dict

if TYPE_CHECKING:
    from typing import Any, TypeVar

    AtomsSchema = TypeVar("AtomsSchema")


def atoms_to_metadata(
    atoms: Atoms,
    charge_and_multiplicity: tuple[int, int] | None = None,
    get_metadata: bool = True,
    strip_info: bool = False,
    store_pmg: bool = True,
    remove_empties: bool = False,
    additional_fields: dict | None = None,
) -> AtomsSchema:
    """
    Convert an ASE Atoms object to a dict suitable for storage in MongoDB.

    Parameters
    ----------
    atoms
        ASE Atoms object to store in {"atoms": atoms}
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    get_metadata
        Whether to store atoms metadata in the returned dict.
    strip_info
        Whether to strip the data from atoms.info in the returned {"atoms":
        atoms}. Note that this data will be stored in {"atoms_info": atoms.info}
        regardless
    store_pmg
        Whether to store the Pymatgen Structure/Molecule object in {"structure":
        Structure} or {"molecule": Molecule}, respectively.
    remove_empties
        Whether to remove None values and empty lists/dicts from the
        TaskDocument.
    additional_fields
        Additional fields to add to the document.

    Returns
    -------
    dict
        Dictionary representation of the atoms object with the following fields:

        - atoms: Atoms = Field(None, title = "The Atoms object from the
          calculation result.")
        - atoms_info: dict = Field(None, title = "The Atoms object info obtained
          from atoms.info.")
        - builder_meta: EmmetMeta = Field(default_factory=EmmetMeta,
          description="Builder metadata."):
            - builder_meta.build_date: str =
              Field(default_factory=datetime.utcnow, description="The build date
              for this document.")
            - builder_meta.database_version: str = Field(None, description="The
              database version for the built data.")
            - builder_meta.emmet_version: str = Field(__version__,
              description="The version of emmet this document was built with.")
            - builder_meta.pymatgen_version: str = Field(pmg_version,
              description="The version of pymatgen this document was built
              with.")
            - builder_meta.pull_request: int = Field(None, description="The pull
              request number associated with this data build.")

        For periodic structures, the task document also has the following
        fields:

        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in the material.")
        - composition: Composition = Field(None, description="Full composition
          for the material.")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified representation of the
          composition.")
        - density: float = Field(None, title="Density", description="Density in
          grams per cm^3.")
        - density_atomic: float = Field(None, title="Packing Density",
          description="The atomic packing density in atoms per cm^3.")
        - elements: List[Element] = Field(None, description="List of elements in
          the material.")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation of the formula.")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the formula.")
        - nelements: int = Field(None, description="Number of elements.")
        - nsites: int = Field(None, description="Total number of sites in the
          structure.")
        - structure: Structure = Field(None, title = "The Pymatgen Structure
          object from the Atoms object, if periodic and store_pmg is True.")
        - symmetry: SymmetryData = Field(None, description="Symmetry data for
          this material.")
            - symmetry.crystal_system: CrystalSystem = Field(None,
              title="Crystal System", description="The crystal system for this
              lattice.")
            - symmetry.number: int = Field(None, title="Space Group Number",
              description="The spacegroup number for the lattice.")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice.")
            - symmetry.symbol: str = Field(None, title="Space Group Symbol",
              description="The spacegroup symbol for the lattice.")
            - symmetry.symprec: float = Field(None, title="Symmetry Finding
              Precision", description="The precision given to spglib to
              determine the symmetry of this lattice.")
        - volume: float = Field(None, title="Volume", description="Total volume
          for this structure in Angstroms^3.")

        For molecules that lack periodicity, the task document also has the
        following fields:

        - charge: int = Field(None, description="Charge of the molecule")
        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in the molecule")
        - composition: Composition = Field(None, description="Full composition
          for the molecule")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified representation of the
          composition")
        - elements: List[Element] = Field(None, description="List of elements in
          the molecule")
        - formula_alphabetical: str = Field(None, title="Alphabetical Formula",
          description="Alphabetical molecular formula")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation of the formula")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the formula.")
        - molecule: Molecule = Field(None, title = "The Pymatgen Molecule object
          from the Atoms object, if not periodic and store_pmg is True.")
        - natoms: int = Field(None, description="Total number of atoms in the
          molecule")
        - nelectrons: int = Field(None, title="Number of electrons",
          description="The total number of electrons for the molecule")
        - nelements: int = Field(None, title="Number of Elements")
        - spin_multiplicity: int = Field(None, description="Spin multiplicity of
          the molecule")
        - symmetry: PointGroupData = Field(None, description="Symmetry data for
          this molecule")
            - symmetry.eigen_tolerance: float = Field(None, title="Interia
              Tensor Eigenvalue Tolerance", description="Tolerance to compare
              eigen values of the inertia tensor.")
            - symmetry.linear: bool = Field(None, title="Molecule Linearity",
              description="Is the molecule linear?")
            - symmetry.matrix_tolerance: float = Field(None, title="Symmetry
              Operation Matrix Element Tolerance" description="Tolerance used to
              generate the full set of symmetry operations of the point group.")
            - symmetry.rotation_number: float = Field(None, title="Rotational
              Symmetry Number", description="Rotational symmetry number for the
              molecule")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice")
            - symmetry.tolerance: float = Field(None, title="Point Group
              Analyzer Tolerance", description="Distance tolerance to consider
              sites as symmetrically equivalent.")
    """

    additional_fields = additional_fields or {}
    atoms = copy_atoms(atoms)
    results = {}

    # Get any charge or multiplicity keys
    if charge_and_multiplicity:
        atoms.charge = charge_and_multiplicity[0]
        atoms.spin_multiplicity = charge_and_multiplicity[1]

    # Strip the dummy atoms, if present
    del atoms[[atom.index for atom in atoms if atom.symbol == "X"]]

    # Get Atoms metadata, if requested. emmet already has built-in tools for
    # generating pymatgen Structure/Molecule metadata, so we'll just use that.
    if get_metadata:
        if atoms.pbc.any():
            struct = AseAtomsAdaptor().get_structure(atoms)
            metadata = StructureMetadata().from_structure(struct).dict()
            if store_pmg:
                results["structure"] = struct
        else:
            mol = AseAtomsAdaptor().get_molecule(atoms, charge_spin_check=False)
            metadata = MoleculeMetadata().from_molecule(mol).dict()
            if store_pmg:
                results["molecule"] = mol
        metadata["builder_meta"]["build_date"] = str(
            metadata["builder_meta"]["build_date"]
        )
    else:
        metadata = {}

    # Copy the info flags as a separate entry in the DB for easy querying
    results["atoms_info"] = _quacc_sanitize(atoms.info)

    # Strip info if requested
    if strip_info:
        atoms_no_info = copy_atoms(atoms)
        atoms_no_info.info = {}
        results["atoms"] = atoms_no_info
    else:
        results["atoms"] = atoms

    # Combine the metadata and results dictionaries
    atoms_doc = metadata | results | additional_fields

    return clean_dict(atoms_doc, remove_empties=remove_empties)


def fetch_atoms(atoms: Atoms | dict) -> Atoms:
    """
    Returns an Atoms object from a typical quacc recipe input, which can either
    be an `Atoms` object or a dictionary with an entry `{"atoms": Atoms}`. It
    may seem a bit silly to do this, but there is a purpose. If you want to
    write a workflow where the output of one recipe is passed to the input of
    another recipe, you can always do `output["atoms"]` to fetch the output
    Atoms object to pass to the input to the second function. However, this
    process will often be its own compute step in workflow management tools
    because they need to resolve the output in order to query it. Depending on
    the workflow manager, this can be a waste of compute resources, so it's
    oftentimes better to do this parsing inside the compute task itself, which
    is why passing in the raw dictionary can be preferred.

    Parameters
    ----------
    atoms
        Atoms object or dictionary with an entry `{"atoms": Atoms}`

    Returns
    -------
    Atoms
        Atoms object
    """
    try:
        return atoms["atoms"]
    except Exception:
        return atoms


def _quacc_sanitize(obj: Any) -> Any:
    """
    Sanitizes an object for storage in MongoDB.

    This is an analogue of monty's jsanitize function but meant to serialize
    Atom/Atoms objects as well.

    Parameters
    ----------
    obj
        Object to sanitize

    Returns
    -------
    Any
        Sanitized object
    """
    if isinstance(obj, (Atom, Atoms)):
        obj = atoms_to_metadata(obj)
    elif isinstance(obj, (list, tuple, np.ndarray)):
        obj = [_quacc_sanitize(i) for i in obj]
    elif isinstance(obj, dict):
        obj = {k.__str__(): _quacc_sanitize(v) for k, v in obj.items()}
    else:
        obj = jsanitize(obj)
    return obj



"""Core recipes for VASP"""
from __future__ import annotations

from typing import TYPE_CHECKING

from quacc import job
from quacc.calculators.vasp import Vasp
from quacc.schemas import fetch_atoms
from quacc.schemas.vasp import summarize_run
from quacc.utils.calc import run_calc
from quacc.utils.dicts import merge_dicts

if TYPE_CHECKING:
    from ase import Atoms

    from quacc.schemas.vasp import VaspSchema

    class DoubleRelaxSchema(VaspSchema):
        relax1: VaspSchema


@job
def static_job(
    atoms: Atoms | dict,
    preset: str | None = None,
    calc_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> VaspSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    preset
        Preset to use.
    calc_swaps
        Dictionary of custom kwargs for the calculator. Overrides the following
        defaults:

        ```python
        {
            "ismear": -5,
            "laechg": True,
            "lcharg": True,
            "lreal": False,
            "lwave": True,
            "nedos": 5001,
            "nsw": 0,
        }
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    VaspSchema
        Dictionary of results from quacc.schemas.vasp.summarize_run
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}

    defaults = {
        "ismear": -5,
        "laechg": True,
        "lcharg": True,
        "lreal": False,
        "lwave": True,
        "nedos": 5001,
        "nsw": 0,
    }
    flags = merge_dicts(defaults, calc_swaps, remove_empties=False)

    atoms.calc = Vasp(atoms, preset=preset, **flags)
    atoms = run_calc(atoms, copy_files=copy_files)

    return summarize_run(atoms, additional_fields={"name": "VASP Static"})


@job
def relax_job(
    atoms: Atoms | dict,
    preset: str | None = None,
    relax_cell: bool = True,
    calc_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> VaspSchema:
    """
    Relax a structure.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    preset
        Preset to use.
    relax_cell
        True if a volume relaxation (ISIF = 3) should be performed. False if
        only the positions (ISIF = 2) should be updated.
    calc_swaps
        Dictionary of custom kwargs for the calculator. Overrides the following
        defaults:

        ```python
        {
            "ediffg": -0.02,
            "isif": 3 if relax_cell else 2,
            "ibrion": 2,
            "isym": 0,
            "lcharg": False,
            "lwave": False,
            "nsw": 200,
        }
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    VaspSchema
        Dictionary of results from quacc.schemas.vasp.summarize_run
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}

    defaults = {
        "ediffg": -0.02,
        "isif": 3 if relax_cell else 2,
        "ibrion": 2,
        "isym": 0,
        "lcharg": False,
        "lwave": False,
        "nsw": 200,
    }
    flags = merge_dicts(defaults, calc_swaps, remove_empties=False)

    atoms.calc = Vasp(atoms, preset=preset, **flags)
    atoms = run_calc(atoms, copy_files=copy_files)

    return summarize_run(atoms, additional_fields={"name": "VASP Relax"})


@job
def double_relax_job(
    atoms: Atoms | dict,
    preset: str | None = None,
    relax_cell: bool = True,
    calc_swaps1: dict | None = None,
    calc_swaps2: dict | None = None,
    copy_files: list[str] | None = None,
) -> DoubleRelaxSchema:
    """
    double_relax a structure. This is particularly useful for a few reasons:

    1. To carry out a cheaper pre-relaxation before the high-quality run.

    2. To carry out a GGA calculation before a meta-GGA or hybrid calculation
    that requires the GGA wavefunction.

    3. To carry out volume relaxations where large changes in volume
    can require a second relaxation to resolve forces.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    preset
        Preset to use.
    relax_cell
        True if a volume relaxation (ISIF = 3) should be performed. False if
        only the positions (ISIF = 2) should be updated.
    calc_swaps1
        Dictionary of custom kwargs for the first relaxation.
    calc_swaps2
        Dictionary of custom kwargs for the second relaxation.
    copy_files
        Files to copy to the (first) runtime directory.

    Returns
    -------
    DoubleRelaxSchema
        Dictionary of results
    """
    atoms = fetch_atoms(atoms)
    calc_swaps1 = calc_swaps1 or {}
    calc_swaps2 = calc_swaps2 or {}

    # Run first relaxation
    summary1 = relax_job.__wrapped__(
        atoms,
        preset=preset,
        relax_cell=relax_cell,
        calc_swaps=calc_swaps1,
        copy_files=copy_files,
    )

    # Run second relaxation
    summary2 = relax_job.__wrapped__(
        summary1,
        preset=preset,
        relax_cell=relax_cell,
        calc_swaps=calc_swaps2,
        copy_files=["WAVECAR"],
    )
    summary2["relax1"] = summary1

    return summary2



"""Core recipes for the tblite code"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ase.optimize import FIRE
from monty.dev import requires

from quacc import job
from quacc.schemas import fetch_atoms
from quacc.schemas.ase import (
    summarize_opt_run,
    summarize_run,
    summarize_thermo,
    summarize_vib_run,
)
from quacc.utils.calc import run_ase_opt, run_ase_vib, run_calc
from quacc.utils.dicts import merge_dicts
from quacc.utils.thermo import ideal_gas

try:
    from tblite.ase import TBLite
except ImportError:
    TBLite = None

if TYPE_CHECKING:
    from typing import Literal

    from ase import Atoms

    from quacc.schemas.ase import OptSchema, RunSchema, ThermoSchema, VibSchema

    class FreqSchema(VibSchema):
        thermo: ThermoSchema


@job
@requires(TBLite, "tblite must be installed. Try pip install tblite[ase]")
def static_job(
    atoms: Atoms | dict,
    method: Literal["GFN1-xTB", "GFN2-xTB", "IPEA1-xTB"] = "GFN2-xTB",
    calc_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> RunSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    method
        GFN1-xTB, GFN2-xTB, and IPEA1-xTB.
    calc_swaps
        Dictionary of custom kwargs for the tblite calculator. Overrides the
        following defaults: `{}`.
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    RunSchema
        Dictionary of results from `quacc.schemas.ase.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}

    atoms.calc = TBLite(method=method, **calc_swaps)
    final_atoms = run_calc(atoms, copy_files=copy_files)
    return summarize_run(
        final_atoms,
        input_atoms=atoms,
        additional_fields={"name": "TBLite Static"},
    )


@job
@requires(TBLite, "tblite must be installed. Try pip install tblite[ase]")
def relax_job(
    atoms: Atoms | dict,
    method: Literal["GFN1-xTB", "GFN2-xTB", "IPEA1-xTB"] = "GFN2-xTB",
    relax_cell: bool = False,
    calc_swaps: dict | None = None,
    opt_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> OptSchema:
    """
    Relax a structure.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    method
        GFN0-xTB, GFN1-xTB, GFN2-xTB.
    relax_cell
        Whether to relax the cell.
    calc_swaps
        Dictionary of custom kwargs for the tblite calculator. Overrides the
        following defaults: `{}`
    opt_swaps
        Dictionary of custom kwargs for `run_ase_opt`. Overrides the following
        defaults:

        ```python
        {"fmax": 0.01, "max_steps": 1000, "optimizer": FIRE}
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    OptSchema
        Dictionary of results from `quacc.schemas.ase.summarize_opt_run`
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}
    opt_swaps = opt_swaps or {}

    opt_defaults = {"fmax": 0.01, "max_steps": 1000, "optimizer": FIRE}
    opt_flags = merge_dicts(opt_defaults, opt_swaps)

    atoms.calc = TBLite(method=method, **calc_swaps)
    dyn = run_ase_opt(atoms, relax_cell=relax_cell, copy_files=copy_files, **opt_flags)

    return summarize_opt_run(dyn, additional_fields={"name": "TBLite Relax"})


@job
@requires(TBLite, "tblite must be installed. Try pip install tblite[ase]")
def freq_job(
    atoms: Atoms | dict,
    method: Literal["GFN1-xTB", "GFN2-xTB", "IPEA1-xTB"] = "GFN2-xTB",
    energy: float = 0.0,
    temperature: float = 298.15,
    pressure: float = 1.0,
    calc_swaps: dict | None = None,
    vib_kwargs: dict | None = None,
    copy_files: list[str] | None = None,
) -> FreqSchema:
    """
    Run a frequency job and calculate thermochemistry.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    method
        GFN0-xTB, GFN1-xTB, GFN2-xTB, GFN-FF.
    energy
        Potential energy in eV. If 0, then the output is just the correction.
    temperature
        Temperature in Kelvins.
    pressure
        Pressure in bar.
    calc_swaps
        dictionary of custom kwargs for the xTB calculator. Overrides the
        following defaults: `{}`
    vib_kwargs
        dictionary of custom kwargs for the Vibrations object.
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    FreqSchema
        Dictionary of results from `quacc.schemas.ase.summarize_vib_run` patched
        with the results of `quacc.schemas.ase.summarize_thermo` in the "thermo"
        key.
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}
    vib_kwargs = vib_kwargs or {}

    atoms.calc = TBLite(method=method, **calc_swaps)
    vibrations = run_ase_vib(atoms, vib_kwargs=vib_kwargs, copy_files=copy_files)
    vib_summary = summarize_vib_run(
        vibrations, additional_fields={"name": "TBLite Frequency Analysis"}
    )

    igt = ideal_gas(atoms, vibrations.get_frequencies(), energy=energy)
    vib_summary["thermo"] = summarize_thermo(
        igt, temperature=temperature, pressure=pressure
    )

    return vib_summary



"""Recipes for slabs"""
from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from quacc import flow, job, subflow
from quacc.calculators.vasp import Vasp
from quacc.schemas import fetch_atoms
from quacc.schemas.vasp import summarize_run
from quacc.utils.calc import run_calc
from quacc.utils.dicts import merge_dicts
from quacc.utils.slabs import make_adsorbate_structures, make_max_slabs_from_bulk

if TYPE_CHECKING:
    from ase import Atoms

    from quacc.schemas.vasp import VaspSchema


@job
def slab_static_job(
    atoms: Atoms | dict,
    preset: str | None = None,
    calc_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> VaspSchema:
    """
    Function to carry out a single-point calculation on a slab.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    preset
        Preset to use.
    calc_swaps
        dictionary of custom kwargs for the calculator. Overrides the following
        defaults:

        ```python
        {
            "auto_dipole": True,
            "ismear": -5,
            "laechg": True,
            "lcharg": True,
            "lreal": False,
            "lvhar": True,
            "lwave": True,
            "nedos": 5001,
            "nsw": 0,
        }
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    VaspSchema
        Dictionary of results from `quacc.schemas.vasp.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}

    defaults = {
        "auto_dipole": True,
        "ismear": -5,
        "laechg": True,
        "lcharg": True,
        "lreal": False,
        "lvhar": True,
        "lwave": True,
        "nedos": 5001,
        "nsw": 0,
    }
    flags = merge_dicts(defaults, calc_swaps, remove_empties=False)

    atoms.calc = Vasp(atoms, preset=preset, **flags)
    atoms = run_calc(atoms, copy_files=copy_files)

    return summarize_run(atoms, additional_fields={"name": "VASP Slab Static"})


@job
def slab_relax_job(
    atoms: Atoms | dict,
    preset: str | None = None,
    calc_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> VaspSchema:
    """
    Function to relax a slab.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    preset
        Preset to use.
    calc_swaps
        Dictionary of custom kwargs for the calculator. Overrides the following
        defaults:

        ```python
        {
            "auto_dipole": True,
            "ediffg": -0.02,
            "isif": 2,
            "ibrion": 2,
            "isym": 0,
            "lcharg": False,
            "lwave": False,
            "nsw": 200,
        }
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    VaspSchema
        Dictionary of results from `quacc.schemas.vasp.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}

    defaults = {
        "auto_dipole": True,
        "ediffg": -0.02,
        "isif": 2,
        "ibrion": 2,
        "isym": 0,
        "lcharg": False,
        "lwave": False,
        "nsw": 200,
    }
    flags = merge_dicts(defaults, calc_swaps, remove_empties=False)

    atoms.calc = Vasp(atoms, preset=preset, **flags)
    atoms = run_calc(atoms, copy_files=copy_files)

    return summarize_run(atoms, additional_fields={"name": "VASP Slab Relax"})


@flow
def bulk_to_slabs_flow(
    atoms: Atoms | dict,
    make_slabs_kwargs: dict | None = None,
    slab_relax: Callable = slab_relax_job,
    slab_static: Callable | None = slab_static_job,
    slab_relax_kwargs: dict | None = None,
    slab_static_kwargs: dict | None = None,
) -> list[VaspSchema]:
    """
    Workflow consisting of:

    1. Slab generation

    2. Slab relaxations

    3. Slab statics (optional)

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    make_slabs_kwargs
        Additional keyword arguments to pass to make_max_slabs_from_bulk()
    slab_relax
        Default to use for the relaxation of the slab structures.
    slab_static
        Default to use for the static calculation of the slab structures.
    slab_relax_kwargs
        Additional keyword arguments to pass to the relaxation calculation.
    slab_static_kwargs
        Additional keyword arguments to pass to the static calculation.

    Returns
    -------
    list[VaspSchema]
        List of dictionary results from `quacc.schemas.vasp.summarize_run`
    """
    slab_relax_kwargs = slab_relax_kwargs or {}
    slab_static_kwargs = slab_static_kwargs or {}
    make_slabs_kwargs = make_slabs_kwargs or {}

    @job
    def _make_slabs(atoms):
        atoms = fetch_atoms(atoms)
        return make_max_slabs_from_bulk(atoms, **make_slabs_kwargs)

    @subflow
    def _relax_distributed(slabs):
        return [slab_relax(slab, **slab_relax_kwargs) for slab in slabs]

    @subflow
    def _relax_and_static_distributed(slabs):
        return [
            slab_static(
                slab_relax(slab, **slab_relax_kwargs),
                **slab_static_kwargs,
            )
            for slab in slabs
        ]

    slabs = _make_slabs(atoms)

    if slab_static is None:
        return _relax_distributed(slabs)

    return _relax_and_static_distributed(slabs)


@flow
def slab_to_ads_flow(
    slab: Atoms,
    adsorbate: Atoms,
    make_ads_kwargs: dict | None = None,
    slab_relax: Callable = slab_relax_job,
    slab_static: Callable | None = slab_static_job,
    slab_relax_kwargs: dict | None = None,
    slab_static_kwargs: dict | None = None,
) -> list[VaspSchema]:
    """
    Workflow consisting of: 1. Slab-adsorbate generation 2. Slab-adsorbate
    relaxations 3. Slab-adsorbate statics (optional)

    Parameters
    ----------
    slab
        Atoms object for the slab structure.
    adsorbate
        Atoms object for the adsorbate.
    make_ads_kwargs
        Additional keyword arguments to pass to make_adsorbate_structures()
    slab_relax
        Default to use for the relaxation of the slab structure.
    slab_static
        Default to use for the static calculation of the slab structures.
    slab_relax_kwargs
        Additional keyword arguments to pass to the relaxation calculation.
    slab_static_kwargs
        Additional keyword arguments to pass to the static calculation.

    Returns
    -------
    list[VaspSchema]
        List of dictionaries of results from quacc.schemas.vasp.summarize_run
    """

    slab_relax_kwargs = slab_relax_kwargs or {}
    slab_static_kwargs = slab_static_kwargs or {}
    make_ads_kwargs = make_ads_kwargs or {}

    @job
    def _make_ads_slabs(atoms, adsorbate):
        atoms = fetch_atoms(atoms)
        return make_adsorbate_structures(atoms, adsorbate, **make_ads_kwargs)

    @subflow
    def _relax_distributed(slabs):
        return [slab_relax(slab, **slab_relax_kwargs) for slab in slabs]

    @subflow
    def _relax_and_static_distributed(slabs):
        return [
            slab_static(
                slab_relax(slab, **slab_relax_kwargs),
                **slab_static_kwargs,
            )
            for slab in slabs
        ]

    ads_slabs = _make_ads_slabs(slab, adsorbate)

    if slab_static is None:
        return _relax_distributed(ads_slabs)

    return _relax_and_static_distributed(ads_slabs)



"""Core recipes for the Q-Chem"""
from __future__ import annotations

import multiprocessing
from typing import TYPE_CHECKING

from ase.optimize import FIRE

from quacc import job
from quacc.calculators.qchem import QChem
from quacc.schemas import fetch_atoms
from quacc.schemas.ase import summarize_opt_run, summarize_run
from quacc.utils.atoms import check_charge_and_spin
from quacc.utils.calc import run_ase_opt, run_calc
from quacc.utils.dicts import merge_dicts, remove_dict_empties

try:
    from sella import Sella

    has_sella = True
except ImportError:
    has_sella = False

if TYPE_CHECKING:
    from ase import Atoms

    from quacc.schemas.ase import OptSchema, RunSchema


@job
def static_job(
    atoms: Atoms | dict,
    charge: int | None = None,
    spin_multiplicity: int | None = None,
    method: str = "wb97mv",
    basis: str = "def2-tzvpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict | None = None,
) -> RunSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    charge
        The total charge of the molecular system. Effectively defaults to zero.
    spin_multiplicity
        The spin multiplicity of the molecular system. Effectively defaults to
        the lowest spin state given the molecular structure and charge.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-TZVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum impicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to pymatgen.io.qchem.QChemDictSet which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.

    Returns
    -------
    RunSchema
        Dictionary of results from `quacc.schemas.ase.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    checked_charge, checked_spin_multiplicity = check_charge_and_spin(
        atoms, charge, spin_multiplicity
    )

    qchem_defaults = {
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or multiprocessing.cpu_count(),
        "qchem_input_params": {
            "basis_set": basis,
            "scf_algorithm": scf_algorithm,
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else None,
        },
    }
    qchem_flags = remove_dict_empties(qchem_defaults)

    atoms.calc = QChem(atoms, **qchem_flags)
    final_atoms = run_calc(atoms)

    return summarize_run(
        final_atoms,
        input_atoms=atoms,
        charge_and_multiplicity=(checked_charge, checked_spin_multiplicity),
        additional_fields={"name": "Q-Chem Static"},
    )


@job
def relax_job(
    atoms: Atoms | dict,
    charge: int | None = None,
    spin_multiplicity: int | None = None,
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict | None = None,
    opt_swaps: dict | None = None,
) -> OptSchema:
    """
    Optimize aka "relax" a molecular structure.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    charge
        The total charge of the molecular system. Effectively defaults to zero.
    spin_multiplicity
        The spin multiplicity of the molecular system. Effectively defaults to
        the lowest spin state given the molecular structure and charge.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-SVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum impicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to pymatgen.io.qchem.QChemDictSet which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    opt_swaps
        Dictionary of custom kwargs for `run_ase_opt`. Overrides the following
        defaults:

        ```python
        {"fmax": 0.01, "max_steps": 1000, "optimizer": "Sella"}
        ```
    Returns
    -------
    OptSchema
        Dictionary of results from `quacc.schemas.ase.summarize_opt_run`
    """

    # TODO: exposing TRICs?
    atoms = fetch_atoms(atoms)
    checked_charge, checked_spin_multiplicity = check_charge_and_spin(
        atoms, charge, spin_multiplicity
    )

    qchem_defaults = {
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or multiprocessing.cpu_count(),
        "qchem_input_params": {
            "basis_set": basis,
            "scf_algorithm": scf_algorithm,
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else None,
        },
    }
    qchem_flags = remove_dict_empties(qchem_defaults)

    opt_swaps = opt_swaps or {}
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": Sella if has_sella else FIRE,
    }
    opt_flags = merge_dicts(opt_defaults, opt_swaps)
    if opt_flags["optimizer"].__name__ == "Sella" and "order" not in opt_flags.get(
        "optimizer_kwargs", {}
    ):
        if "optimizer_kwargs" not in opt_flags:
            opt_flags["optimizer_kwargs"] = {}
        opt_flags["optimizer_kwargs"]["order"] = 0

    atoms.calc = QChem(atoms, **qchem_flags)
    dyn = run_ase_opt(atoms, **opt_flags)

    return summarize_opt_run(
        dyn,
        charge_and_multiplicity=(checked_charge, checked_spin_multiplicity),
        additional_fields={"name": "Q-Chem Optimization"},
    )



"""
Transition state recipes for the NewtonNet code
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ase.optimize import FIRE
from monty.dev import requires

from quacc import SETTINGS, job
from quacc.recipes.newtonnet.core import _add_stdev_and_hess
from quacc.recipes.newtonnet.core import freq_job as _freq_job
from quacc.recipes.newtonnet.core import relax_job
from quacc.schemas import fetch_atoms
from quacc.schemas.ase import summarize_opt_run
from quacc.utils.calc import run_ase_opt
from quacc.utils.dicts import merge_dicts

try:
    from sella import IRC, Sella
except ImportError:
    Sella = None

try:
    from newtonnet.utils.ase_interface import MLAseCalculator as NewtonNet
except ImportError:
    NewtonNet = None

if TYPE_CHECKING:
    from typing import Literal

    import numpy as np
    from ase import Atoms

    from quacc.recipes.newtonnet.core import FreqSchema
    from quacc.schemas.ase import OptSchema
    from quacc.utils.wflows import Job

    class TSSchema(OptSchema):
        freq: FreqSchema | None

    class IRCSchema(OptSchema):
        freq: FreqSchema | None

    class QuasiIRCSchema(OptSchema):
        irc: IRCSchema
        freq: FreqSchema | None


@job
@requires(NewtonNet, "NewtonNet must be installed. Try pip install quacc[newtonnet]")
@requires(Sella, "Sella must be installed. Try pip install quacc[optimizers]")
def ts_job(
    atoms: Atoms | dict,
    use_custom_hessian: bool = False,
    freq_job: Job | None = _freq_job,
    freq_job_kwargs: dict | None = None,
    calc_swaps: dict | None = None,
    opt_swaps: dict | None = None,
) -> TSSchema:
    """
    Perform a transition state (TS) job using the given atoms object.

    Parameters
    ----------
    atoms
        The atoms object representing the system.
    use_custom_hessian
        Whether to use a custom Hessian matrix.
    freq_job
        Default Job to use for the frequency analysis.
    freq_job_kwargs
        Keyword arguments to use for the `freq_job`.
    calc_swaps
        Optional swaps for the NewtonNet calculator. Overrides the
        following defaults:

        ```python
        {
            "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
            "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
        }
        ```
    opt_swaps
        Optional swaps for the optimization parameters. Overrides the
        following defaults:

        ```python
        {
            "fmax": 0.01,
            "max_steps": 1000,
            "optimizer": Sella,
            "optimizer_kwargs": {"diag_every_n": 0} if use_custom_hessian else {},
        }
        ```

    Returns
    -------
    TSSchema
        Dictionary of results, specified in `quacc.schemas.ase.RunSchema`
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}
    opt_swaps = opt_swaps or {}
    freq_job_kwargs = freq_job_kwargs or {}

    defaults = {
        "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
        "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
    }
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": Sella,
        "optimizer_kwargs": {"diag_every_n": 0} if use_custom_hessian else {},
    }

    flags = merge_dicts(defaults, calc_swaps)
    opt_flags = merge_dicts(opt_defaults, opt_swaps)

    atoms.calc = NewtonNet(**flags)

    if use_custom_hessian:
        if opt_flags.get("optimizer", FIRE).__name__ != "Sella":
            raise ValueError("Custom hessian can only be used with Sella.")

        opt_flags["optimizer_kwargs"]["hessian_function"] = _get_hessian

    ml_calculator = NewtonNet(**flags)
    atoms.calc = ml_calculator

    # Run the TS optimization
    dyn = run_ase_opt(atoms, **opt_flags)
    opt_ts_summary = _add_stdev_and_hess(
        summarize_opt_run(dyn, additional_fields={"name": "NewtonNet TS"})
    )

    # Run a frequency calculation
    freq_summary = (
        freq_job.__wrapped__(opt_ts_summary, **freq_job_kwargs) if freq_job else None
    )
    opt_ts_summary["freq"] = freq_summary

    return opt_ts_summary


@job
@requires(NewtonNet, "NewtonNet must be installed. Try pip install quacc[newtonnet]")
@requires(Sella, "Sella must be installed. Try pip install quacc[optimizers]")
def irc_job(
    atoms: Atoms | dict,
    direction: Literal["forward", "reverse"] = "forward",
    freq_job: Job | None = _freq_job,
    freq_job_kwargs: dict | None = None,
    calc_swaps: dict | None = None,
    opt_swaps: dict | None = None,
) -> IRCSchema:
    """
    Perform an intrinsic reaction coordinate (IRC) job using the given atoms
    object.

    Parameters
    ----------
    atoms
        The atoms object representing the system.
    direction
        The direction of the IRC calculation ("forward" or "reverse").
    freq_job
        Default Job to use for the frequency analysis.
    freq_job_kwargs
        Keyword arguments for the `freq_job`.
    calc_swaps
        Optional swaps for the calculator. Overrides the following
        defaults:

        ```python
        {
            "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
            "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
        }
        ```
    opt_swaps
        Optional swaps for the optimization parameters. Overrides the
        following defaults:

        ```python
        {
            "fmax": 0.01,
            "max_steps": 1000,
            "optimizer": IRC,
            "optimizer_kwargs": {
                "dx": 0.1,
                "eta": 1e-4,
                "gamma": 0.4,
                "keep_going": True,
            },
            "run_kwargs": {
                "direction": direction,
            },
        }
        ```

    Returns
    -------
    IRCSchema
        A dictionary containing the IRC summary and thermodynamic summary.
    """
    atoms = fetch_atoms(atoms)
    calc_swaps = calc_swaps or {}
    opt_swaps = opt_swaps or {}
    freq_job_kwargs = freq_job_kwargs or {}
    default_settings = SETTINGS.copy()

    defaults = {
        "model_path": SETTINGS.NEWTONNET_MODEL_PATH,
        "settings_path": SETTINGS.NEWTONNET_CONFIG_PATH,
    }
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": IRC,
        "optimizer_kwargs": {
            "dx": 0.1,
            "eta": 1e-4,
            "gamma": 0.4,
            "keep_going": True,
        },
        "run_kwargs": {
            "direction": direction,
        },
    }

    flags = merge_dicts(defaults, calc_swaps)
    opt_flags = merge_dicts(opt_defaults, opt_swaps)

    # Define calculator
    atoms.calc = NewtonNet(**flags)

    # Run IRC
    SETTINGS.CHECK_CONVERGENCE = False
    dyn = run_ase_opt(atoms, **opt_flags)
    opt_irc_summary = _add_stdev_and_hess(
        summarize_opt_run(
            dyn, additional_fields={"name": f"NewtonNet IRC: {direction}"}
        )
    )
    SETTINGS.CHECK_CONVERGENCE = default_settings.CHECK_CONVERGENCE

    # Run frequency job
    freq_summary = (
        freq_job.__wrapped__(opt_irc_summary, **freq_job_kwargs) if freq_job else None
    )
    opt_irc_summary["freq"] = freq_summary

    return opt_irc_summary


@job
@requires(NewtonNet, "NewtonNet must be installed. Try pip install quacc[newtonnet]")
@requires(Sella, "Sella must be installed. Try pip install quacc[optimizers]")
def quasi_irc_job(
    atoms: Atoms | dict,
    direction: Literal["forward", "reverse"] = "forward",
    freq_job: Job | None = _freq_job,
    freq_job_kwargs: dict | None = None,
    irc_swaps: dict | None = None,
    opt_swaps: dict | None = None,
) -> QuasiIRCSchema:
    """
    Perform a quasi-IRC job using the given atoms object.

    Parameters
    ----------
    atoms
        The atoms object representing the system.
    direction
        The direction of the IRC calculation ("forward" or "reverse").
    freq_job
        Default Job to use for the frequency analysis.
    freq_job_kwargs
        Keyword arguments for `freq_job`.
    irc_swaps
        Optional swaps for the IRC optimization parameters. Overrides
        the following defaults: `{"max_steps": 5}`.
    opt_swaps
        Optional swaps for the optimization parameters. Overrides
        the following defaults: `{}`.

    Returns
    -------
    QuasiIRCSchema
        A dictionary containing the IRC summary, optimization summary, and
        thermodynamic summary.
    """
    irc_swaps = irc_swaps or {}
    opt_swaps = opt_swaps or {}
    freq_job_kwargs = freq_job_kwargs or {}

    irc_defaults = {"max_steps": 5}
    irc_flags = merge_dicts(irc_defaults, irc_swaps)

    # Run IRC
    irc_summary = irc_job.__wrapped__(
        atoms, direction=direction, opt_swaps=irc_flags, freq_job=None
    )

    # Run opt
    relax_summary = relax_job.__wrapped__(irc_summary, **opt_swaps)

    # Run frequency
    freq_summary = (
        freq_job.__wrapped__(relax_summary, **freq_job_kwargs) if freq_job else None
    )
    relax_summary["freq"] = freq_summary
    relax_summary["irc"] = irc_summary

    return relax_summary


def _get_hessian(atoms: Atoms) -> np.ndarray:
    """
    Calculate and retrieve the Hessian matrix for the given molecular
    configuration.

    This function takes an ASE Atoms object representing a molecular
    configuration and uses the NewtonNet machine learning calculator to
    calculate the Hessian matrix. The calculated Hessian matrix is then
    returned.

    Parameters
    ----------
    atoms
        The ASE Atoms object representing the molecular configuration.

    Returns
    -------
    np.ndarray
        The calculated Hessian matrix, reshaped into a 2D array.
    """
    ml_calculator = NewtonNet(
        model_path=SETTINGS.NEWTONNET_MODEL_PATH,
        settings_path=SETTINGS.NEWTONNET_CONFIG_PATH,
    )
    ml_calculator.calculate(atoms)

    return ml_calculator.results["hessian"].reshape((-1, 3 * len(atoms)))



"""
Utility functions for dealing with defects
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from pymatgen.analysis.defects.generators import VacancyGenerator
from pymatgen.analysis.defects.thermo import DefectEntry
from pymatgen.core.periodic_table import DummySpecies
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.io.ase import AseAtomsAdaptor
from shakenbreak.input import Distortions

if TYPE_CHECKING:
    import numpy as np
    from ase.atoms import Atoms
    from pymatgen.analysis.defects.core import Defect
    from pymatgen.analysis.defects.generators import (
        AntiSiteGenerator,
        ChargeInterstitialGenerator,
        InterstitialGenerator,
        SubstitutionGenerator,
        VoronoiInterstitialGenerator,
    )
    from pymatgen.core import Structure


def make_defects_from_bulk(
    atoms: Atoms,
    defect_gen: (
        AntiSiteGenerator
        | ChargeInterstitialGenerator
        | InterstitialGenerator
        | SubstitutionGenerator
        | VacancyGenerator
        | VoronoiInterstitialGenerator
    ) = VacancyGenerator,
    defect_charge: int = 0,
    sc_mat: np.ndarray | None = None,
    min_atoms: int = 80,
    max_atoms: int = 240,
    min_length: float = 10.0,
    force_diagonal: bool = False,
    **defect_gen_kwargs,
) -> list[Atoms]:
    """
    Function to make defects from a bulk atoms object.

    Parameters
    ----------
    atoms
        bulk atoms
    defect_gen
        defect generator
    sc_mat
        supercell matrix
    min_atoms
        minimum number of atoms in supercell
    max_atoms
        maximum number of atoms in supercell
    min_length
        minimum length of supercell
    force_diagonal
        force supercell to be diagonal
    defect_charge
        charge state of defect
    **defect_gen_kwargs
        keyword arguments to pass to the pymatgen.analysis.defects.generators
        get_defects() method

    Returns
    -------
    list[Atoms]
        All generated defects
    """

    # Use pymatgen-analysis-defects and ShakeNBreak to generate defects
    struct = AseAtomsAdaptor.get_structure(atoms)

    # Make all the defects
    defects = defect_gen().get_defects(struct, **defect_gen_kwargs)
    final_defects = []
    for defect in defects:
        defect.user_charges = [defect_charge]

        # Generate the supercell for a defect
        defect_supercell = defect.get_supercell_structure(
            sc_mat=sc_mat,
            dummy_species=DummySpecies(),
            min_atoms=min_atoms,
            max_atoms=max_atoms,
            min_length=min_length,
            force_diagonal=force_diagonal,
        )

        # Generate DefectEntry object from Defect object
        defect_entry = _get_defect_entry_from_defect(
            defect=defect,
            defect_supercell=defect_supercell,
            defect_charge=defect_charge,
        )

        # Instantiate class to apply rattle and bond distortion to all defects
        Dist = Distortions([defect_entry])

        # Apply rattle and bond distortion to all defects
        defect_dict, distortion_metadata = Dist.apply_distortions()
        defect_symbol = next(iter(distortion_metadata["defects"].keys()))
        distortion_dict = defect_dict[defect_symbol]["charges"][defect_charge][
            "structures"
        ]["distortions"]

        # Make atoms objects and store defect stats
        for distortions, defect_struct in distortion_dict.items():
            final_defect = AseAtomsAdaptor.get_atoms(defect_struct)
            defect_stats = {
                "defect_symbol": defect_symbol,
                "defect_charge": defect_charge,
                "distortions": distortions,
                "bulk": atoms,
                "defect": defect,
            }
            final_defect.info["defect_stats"] = defect_stats
            final_defects.append(final_defect)
    return final_defects


def _get_defect_entry_from_defect(
    defect: Defect,
    defect_supercell: Structure,
    defect_charge: int,
) -> DefectEntry:
    """
    Function to generate DefectEntry object from Defect object

    Parameters
    ----------
    defect
        defect object
    defect_supercell
        defect supercell
    defect_charge
        charge state of defect

    Returns
    -------
    DefectEntry
        defect entry
    """
    # Find defect's fractional coordinates and remove it from supercell
    dummy_site = next(
        site
        for site in defect_supercell
        if site.species.elements[0].symbol == DummySpecies().symbol
    )
    sc_defect_frac_coords = dummy_site.frac_coords
    defect_supercell.remove(dummy_site)

    computed_structure_entry = ComputedStructureEntry(
        structure=defect_supercell,
        energy=0.0,  # needs to be set, so set to 0.0
    )

    return DefectEntry(
        defect=defect,
        charge_state=defect_charge,
        sc_entry=computed_structure_entry,
        sc_defect_frac_coords=sc_defect_frac_coords,
    )



"""Core recipes for GULP"""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from ase.calculators.gulp import GULP

from quacc import job
from quacc.schemas import fetch_atoms
from quacc.schemas.ase import summarize_run
from quacc.utils.calc import run_calc
from quacc.utils.dicts import merge_dicts

if TYPE_CHECKING:
    from ase import Atoms

    from quacc.schemas.ase import RunSchema

GEOM_FILE_PBC = "gulp.cif"
GEOM_FILE_NOPBC = "gulp.xyz"


@job
def static_job(
    atoms: Atoms | dict,
    use_gfnff: bool = True,
    library: str | None = None,
    keyword_swaps: dict | None = None,
    option_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> RunSchema:
    """
    Carry out a single-point calculation. Note: 'Conditions' are not yet
    natively supported.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    use_gfnff
        True if (p)GFN-FF should be used; False if not.
    library
        Filename of the potential library file, if required.
    keyword_swaps
        Dictionary of custom keyword swap kwargs for the calculator. Overrides
        the following defaults:

        ```python
        {
            "gfnff": True if use_gfnff else None,
            "gwolf": True if use_gfnff and atoms.pbc.any() else None,
        }
        ```
    option_swaps
        Dictionary of custom option swap kwargs for the calculator. Overrides
        the following defaults:

        ```python
        {
            "dump every gulp.res": True,
            f"output cif {GEOM_FILE_PBC}": True if atoms.pbc.any() else None,
            f"output xyz {GEOM_FILE_NOPBC}": None if atoms.pbc.any() else True,
        }
        ```

    Returns
    -------
    RunSchema
        Dictionary of results from `quacc.schemas.ase.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    keyword_swaps = keyword_swaps or {}
    option_swaps = option_swaps or {}

    default_keywords = {
        "gfnff": True if use_gfnff else None,
        "gwolf": True if use_gfnff and atoms.pbc.any() else None,
    }
    default_options = {
        "dump every gulp.res": True,
        f"output cif {GEOM_FILE_PBC}": True if atoms.pbc.any() else None,
        f"output xyz {GEOM_FILE_NOPBC}": None if atoms.pbc.any() else True,
    }

    keywords = merge_dicts(default_keywords, keyword_swaps)
    options = merge_dicts(default_options, option_swaps)

    gulp_keywords = " ".join(list(keywords.keys()))
    gulp_options = list(options.keys())

    atoms.calc = GULP(keywords=gulp_keywords, options=gulp_options, library=library)
    final_atoms = run_calc(
        atoms,
        geom_file=GEOM_FILE_PBC if atoms.pbc.any() else GEOM_FILE_NOPBC,
        copy_files=copy_files,
    )

    return summarize_run(
        final_atoms,
        input_atoms=atoms,
        additional_fields={"name": "GULP Static"},
    )


@job
def relax_job(
    atoms: Atoms | dict,
    use_gfnff: bool = True,
    library: str | None = None,
    relax_cell: bool = False,
    keyword_swaps: dict | None = None,
    option_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> RunSchema:
    """
    Carry out a single-point calculation. Note: 'Conditions' are not yet
    natively supported.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    use_gfnff
        True if (p)GFN-FF should be used; False if not.
    library
        Filename of the potential library file, if required.
    relax_cell
        True if the volume should be relaxed; False if not.
    keyword_swaps
        Dictionary of custom keyword swap kwargs for the calculator. Overrides
        the following defaults:

        ```python
        {
            "opti": True,
            "gfnff": True if use_gfnff else None,
            "gwolf": True if use_gfnff and atoms.pbc.any() else None,
            "conp": True if relax_cell and atoms.pbc.any() else None,
            "conv": None if relax_cell and atoms.pbc.any() else True,
        }
        ```
    option_swaps
        Dictionary of custom option swap kwargs for the calculator. Overrides
        the following defaults:

        ```python
        {
            "dump every gulp.res": True,
            f"output cif {GEOM_FILE_PBC}": True if atoms.pbc.any() else None,
            f"output xyz {GEOM_FILE_NOPBC}": None if atoms.pbc.any() else True,
        }
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    dict
        Dictionary of results from `quacc.schemas.ase.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    keyword_swaps = keyword_swaps or {}
    option_swaps = option_swaps or {}

    if relax_cell and not atoms.pbc.any():
        warnings.warn(
            "Volume relaxation requested but no PBCs found. Ignoring.", UserWarning
        )
        relax_cell = False

    default_keywords = {
        "opti": True,
        "gfnff": True if use_gfnff else None,
        "gwolf": True if use_gfnff and atoms.pbc.any() else None,
        "conp": True if relax_cell and atoms.pbc.any() else None,
        "conv": None if relax_cell and atoms.pbc.any() else True,
    }
    default_options = {
        "dump every gulp.res": True,
        f"output cif {GEOM_FILE_PBC}": True if atoms.pbc.any() else None,
        f"output xyz {GEOM_FILE_NOPBC}": None if atoms.pbc.any() else True,
    }

    keywords = merge_dicts(default_keywords, keyword_swaps)
    options = merge_dicts(default_options, option_swaps)

    gulp_keywords = " ".join(list(keywords.keys()))
    gulp_options = list(options.keys())

    atoms.calc = GULP(keywords=gulp_keywords, options=gulp_options, library=library)
    final_atoms = run_calc(
        atoms,
        geom_file=GEOM_FILE_PBC if atoms.pbc.any() else GEOM_FILE_NOPBC,
        copy_files=copy_files,
    )

    if not final_atoms.calc.get_opt_state():
        msg = "Optimization did not converge."
        raise ValueError(msg)

    return summarize_run(
        final_atoms,
        input_atoms=atoms,
        additional_fields={"name": "GULP Relax"},
    )



"""Custodian handlers for QChem"""
from __future__ import annotations

import sys

from monty.dev import requires

from quacc import SETTINGS

try:
    import openbabel as ob
except ImportError:
    ob = None


@requires(
    ob,
    "Openbabel must be installed. Try conda install -c conda-forge openbabel",
)
def run_custodian(
    qchem_cores: int = 1,
    qchem_cmd: str | None = None,
    qchem_local_scratch: str | None = None,
    qchem_use_error_handlers: bool | None = None,
    qchem_custodian_max_errors: int | None = None,
) -> None:
    """
    Function to run QChem Custodian

    Parameters
    ----------
    qchem_cores
        Number of cores to use for the Q-Chem calculation.
    qchem_cmd
        Q-Chem command. Defaults to "qchem" in settings.
    qchem_local_scratch
        Compute-node local scratch directory in which Q-Chem should perform IO.
        Defaults to /tmp in settings.
    qchem_use_error_handlers
        Whether or not to employ error handlers. Defaults to True in settings.
    qchem_custodian_max_errors
        Maximum number of errors to allow before stopping the run. Defaults to 5
        in settings.

    Returns
    -------
    None
    """
    # Adapted from atomate.qchem.firetasks.run_calc
    from custodian import Custodian
    from custodian.qchem.handlers import QChemErrorHandler
    from custodian.qchem.jobs import QCJob

    # Set defaults
    qchem_cmd = SETTINGS.QCHEM_CMD if qchem_cmd is None else qchem_cmd
    qchem_local_scratch = (
        SETTINGS.QCHEM_LOCAL_SCRATCH
        if qchem_local_scratch is None
        else qchem_local_scratch
    )
    qchem_use_error_handlers = (
        SETTINGS.QCHEM_USE_ERROR_HANDLERS
        if qchem_use_error_handlers is None
        else qchem_use_error_handlers
    )
    qchem_custodian_max_errors = (
        SETTINGS.QCHEM_CUSTODIAN_MAX_ERRORS
        if qchem_custodian_max_errors is None
        else qchem_custodian_max_errors
    )

    # Error handlers for Q-Chem
    handlers = [QChemErrorHandler()] if qchem_use_error_handlers else []

    # Run Q-Chem
    jobs = [
        QCJob(
            qchem_command=qchem_cmd,
            max_cores=qchem_cores,
            calc_loc=qchem_local_scratch,
        )
    ]

    c = Custodian(
        handlers,
        jobs,
        max_errors=qchem_custodian_max_errors,
    )

    c.run()


if __name__ == "__main__":
    run_custodian(qchem_cores=sys.argv[1]) if len(sys.argv) > 1 else run_custodian()



"""Settings for quacc"""
from __future__ import annotations

import os
from pathlib import Path
from shutil import which
from typing import List, Optional, Union

from pydantic import BaseSettings, Field, root_validator

from quacc.presets import vasp as vasp_defaults

try:
    import covalent
except ImportError:
    covalent = None
try:
    import parsl
except ImportError:
    parsl = None
try:
    import jobflow
except ImportError:
    jobflow = None
try:
    import redun
except ImportError:
    redun = None

_DEFAULT_CONFIG_FILE_PATH = Path("~", ".quacc.yaml").expanduser()


class QuaccSettings(BaseSettings):
    """
    Settings for quacc.

    The default way to modify these is to make a ~/.quacc.yaml file.
    Alternatively, the environment variable QUACC_CONFIG_FILE can be set to
    point to a yaml file with quacc settings.

    The variables can also be modified individually though environment variables
    by using the "QUACC" prefix. e.g. QUACC_SCRATCH_DIR=/path/to/scratch.
    """

    # --8<-- [start:settings]

    # ---------------------------
    # Workflow Engine
    # ---------------------------

    WORKFLOW_ENGINE: str = Field(
        "covalent"
        if covalent
        else "parsl"
        if parsl
        else "redun"
        if redun
        else "jobflow"
        if jobflow
        else "local",
        description=(
            "The workflow manager to use."
            "Options include: 'covalent', 'parsl', 'redun', 'jobflow', or 'local'"
        ),
    )

    # ---------------------------
    # General Settings
    # ---------------------------

    CONFIG_FILE: Union[str, Path] = Field(
        _DEFAULT_CONFIG_FILE_PATH,
        description=(
            "Path to the YAML file to load alternative quacc configuration "
            "defaults from."
        ),
    )
    RESULTS_DIR: Union[str, Path] = Field(
        Path.cwd(),
        description=(
            "Directory to store I/O-based calculation results in."
            "Note that this behavior may be modified by the chosen workflow engine."
            "For instance, Covalent specifies the base directory as the `workdir` "
            "of a local executor or the `remote_workdir` of a remote executor."
            "In this case, the `RESULTS_DIR` will be a subdirectory of that directory."
        ),
    )
    SCRATCH_DIR: Union[str, Path] = Field(
        Path("/tmp") if Path("/tmp").exists() else Path.cwd(),
        description="Scratch directory for calculations.",
    )
    CREATE_UNIQUE_WORKDIR: bool = Field(
        False,
        description=(
            "Whether to have a unique working directory in RESULTS_DIR for each job."
            "Some workflow engines have an option to do this for you already."
        ),
    )
    GZIP_FILES: bool = Field(
        True, description="Whether generated files should be gzip'd."
    )
    CHECK_CONVERGENCE: bool = Field(
        True,
        description="Whether to check for convergence in the `summarize_run`-type functions, if supported.",
    )

    # ---------------------------
    # Data Store Settings
    # ---------------------------
    PRIMARY_STORE: str = Field(
        None,
        description=(
            "String-based JSON representation of the primary Maggma data store "
            "where calculation results will be stored."
            "Taken from the `.to_json()` method of the corresponding Store object."
        ),
    )

    # ---------------------------
    # ORCA Settings
    # ---------------------------
    ORCA_CMD: Union[str, Path] = Field(
        "orca",
        description=(
            "Path to the ORCA executable. This must be the full, absolute path "
            "for parallel calculations to work."
        ),
    )

    # ---------------------------
    # VASP Settings
    # ---------------------------

    # VASP Settings: Main
    VASP_PARALLEL_CMD: str = Field(
        "",
        description=(
            "Parallel command to run VASP with Custodian."
            "For example: srun -N 2 --ntasks-per-node 48"
            "Note that this does not include the executable name."
        ),
    )
    VASP_CMD: str = Field(
        "vasp_std", description="Command to run the standard version of VASP."
    )
    VASP_GAMMA_CMD: str = Field(
        "vasp_gam", description="Command to run the gamma-point only version of VASP."
    )

    # VASP Settings: General
    VASP_INCAR_COPILOT: bool = Field(
        True,
        description=(
            "Whether co-pilot mode should be used for VASP INCAR handling."
            "This will modify INCAR flags on-the-fly if they disobey the VASP manual."
            "A warning will be raised in each case."
        ),
    )
    VASP_BADER: bool = Field(
        bool(which("bader")),
        description=(
            "Whether to run a Bader analysis when summarizing VASP results."
            "Requires bader to be in PATH."
        ),
    )
    VASP_PRESET_MAG_DEFAULT: float = Field(
        1.0,
        description=(
            "Default initial magmom to use for a given element if a preset "
            "with magmoms is provided but an element is missing from the list"
        ),
    )
    VASP_MAG_CUTOFF: float = Field(
        0.05,
        description=(
            "If the absolute value of all magnetic moments are below this value, "
            "they will be set to 0 such that a spin-unpolarized calculation will be performed"
        ),
    )
    VASP_COPY_MAGMOMS: bool = Field(
        True,
        description=(
            "If True, any pre-existing atoms.get_magnetic_moments() will be set"
            "in atoms.set_initial_magnetic_moments()."
        ),
    )
    VASP_VERBOSE: bool = Field(
        True,
        description="If True, warnings will be raised when INCAR parameters are changed.",
    )
    VASP_PRESET_DIR: str = Field(
        os.path.dirname(vasp_defaults.__file__),
        description="Path to the VASP preset directory",
    )

    # VASP Settings: Custodian
    VASP_USE_CUSTODIAN: bool = Field(
        True, description="Whether Custodian should be used to run VASP"
    )
    VASP_CUSTODIAN_VTST: bool = Field(
        False,
        description=(
            "If VTST-related input swaps should be used when running Custodian."
            "Requires VASP to be compiled with VTST"
        ),
    )
    VASP_CUSTODIAN_MAX_ERRORS: int = Field(
        5, description="Maximum errors for Custodian"
    )
    VASP_CUSTODIAN_HANDLERS: List[str] = Field(
        [
            "VaspErrorHandler",
            "MeshSymmetryErrorHandler",
            "UnconvergedErrorHandler",
            "NonConvergingErrorHandler",
            "PotimErrorHandler",
            "PositiveEnergyErrorHandler",
            "FrozenJobErrorHandler",
            "StdErrHandler",
            "LargeSigmaHandler",
            "IncorrectSmearingHandler",
        ],
        description="Handlers for Custodian",
    )
    VASP_CUSTODIAN_VALIDATORS: List[str] = Field(
        ["VasprunXMLValidator", "VaspFilesValidator"],
        description="Validators for Custodian",
    )
    VASP_CUSTODIAN_WALL_TIME: Optional[int] = Field(
        None,
        description=(
            "After this many seconds, Custodian will stop running "
            "and ensure that VASP writes a STOPCAR"
        ),
    )

    # ---------------------------
    # Q-Chem Settings
    # ---------------------------

    # Q-Chem Settings: Main
    QCHEM_CMD: str = Field(
        "qchem", description="Command to run the standard version of Q-Chem."
    )

    QCHEM_LOCAL_SCRATCH: Union[str, Path] = Field(
        Path("/tmp") if Path("/tmp").exists() else Path.cwd(),
        description="Compute-node local scratch directory in which Q-Chem should perform IO.",
    )

    # Q-Chem Settings: Custodian
    QCHEM_USE_ERROR_HANDLERS: bool = Field(
        True,
        description="Whether Custodian's error handlers should be employed for Q-Chem.",
    )

    QCHEM_CUSTODIAN_MAX_ERRORS: int = Field(
        5, description="Maximum errors for Q-Chem Custodian."
    )

    # ---------------------------
    # NewtonNet Settings
    # ---------------------------
    NEWTONNET_MODEL_PATH: Union[Union[str, Path], List[Union[str, Path]]] = Field(
        "best_model_state.tar", description="Path to NewtonNet .tar model"
    )
    NEWTONNET_CONFIG_PATH: Union[Union[str, Path], List[Union[str, Path]]] = Field(
        "config.yml", description="Path to NewtonNet YAML settings file"
    )

    # --8<-- [end:settings]

    class Config:
        """Pydantic config settings."""

        env_prefix = "quacc_"

    @root_validator(pre=True)
    def load_default_settings(cls, values: dict) -> dict:
        """
        Loads settings from a root file if available and uses that as defaults
        in place of built in defaults.

        Parameters
        ----------
        values
            Settings to load.

        Returns
        -------
        dict
            Loaded settings.
        """

        from monty.serialization import loadfn

        config_file_path = Path(
            values.get("CONFIG_FILE", _DEFAULT_CONFIG_FILE_PATH)
        ).expanduser()

        new_values = {}
        if config_file_path.exists() and config_file_path.stat().st_size > 0:
            new_values |= loadfn(config_file_path)

        new_values.update(values)
        return new_values



"""Schemas for storing ASE-based data"""
from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from ase import units
from ase.constraints import Filter
from ase.io import read
from ase.vibrations.data import VibrationsData

from quacc import SETTINGS
from quacc.schemas.atoms import atoms_to_metadata
from quacc.utils.atoms import prep_next_run as prep_next_run_
from quacc.utils.db import results_to_db
from quacc.utils.dicts import clean_dict
from quacc.utils.files import get_uri

if TYPE_CHECKING:
    from ase import Atoms
    from ase.io import Trajectory
    from ase.optimize.optimize import Optimizer
    from ase.thermochemistry import IdealGasThermo
    from ase.vibrations import Vibrations
    from maggma.core import Store
    from tyipng import TypeVar

    RunSchema = TypeVar("RunSchema")
    OptSchema = TypeVar("OptSchema")
    VibSchema = TypeVar("VibSchema")
    ThermoSchema = TypeVar("ThermoSchema")


def summarize_run(
    atoms: Atoms,
    input_atoms: Atoms | None = None,
    charge_and_multiplicity: tuple[int, int] | None = None,
    prep_next_run: bool = True,
    remove_empties: bool = False,
    additional_fields: dict | None = None,
    store: Store | None = None,
) -> RunSchema:
    """
    Get tabulated results from an Atoms object and calculator and store them in
    a database-friendly format. This is meant to be compatible with all
    calculator types.

    Parameters
    ----------
    atoms
        ASE Atoms following a calculation. A calculator must be attached.
    input_atoms
        Input ASE Atoms object to store.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    prep_next_run
        Whether the Atoms object stored in {"atoms": atoms} should be prepared
        for the next run This clears out any attached calculator and moves the
        final magmoms to the initial magmoms.
    remove_empties
        Whether to remove None values and empty lists/dicts from the task
        document.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    dict
        Dictionary representation of the task document with the following
        fields:

        - atoms: Atoms = Field(None, title = "The Atoms object from the
          calculation result.")
        - atoms_info: dict = Field(None, title = "The Atoms object info obtained
          from atoms.info.")
        - builder_meta: EmmetMeta = Field(default_factory=EmmetMeta,
          description="Builder metadata."):
            - builder_meta.build_date: str =
              Field(default_factory=datetime.utcnow, description="The build date
              for this document.")
            - builder_meta.database_version: str = Field(None, description="The
              database version for the built data.")
            - builder_meta.emmet_version: str = Field(__version__,
              description="The version of emmet this document was built with.")
            - builder_meta.pymatgen_version: str = Field(pmg_version,
              description="The version of pymatgen this document was built
              with.")
            - builder_meta.pull_request: int = Field(None, description="The pull
              request number associated with this data build.")
        - dir_name: str = Field(None, description="Directory where the output is
          parsed")
        - input_structure: Molecule | Structure = Field(None, title = "The
          Pymatgen Structure or Molecule object from
         the input Atoms object if input_atoms is not None.")
        - nid: str = Field(None, title = "The node ID representing the machine
          where the calculation was run.")
        - parameters: dict = Field(None, title = "the parameters used to run the
          calculation.")
        - results: dict = Field(None, title = "The results from the
          calculation.")

        For periodic structures, the task document also has the following
        fields: - chemsys: str = Field(None, title="Chemical System",
        description="dash-delimited string of elements in the material.") -
        composition: Composition = Field(None, description="Full composition for
        the material.") - composition_reduced: Composition = Field(None,
        title="Reduced Composition", description="Simplified
         representation of the composition.")
        - density: float = Field(None, title="Density", description="Density in
          grams per cm^3.")
        - density_atomic: float = Field(None, title="Packing Density",
          description="The atomic packing density in
         atoms per cm^3.")
        - elements: List[Element] = Field(None, description="List of elements in
          the material.")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation
         of the formula.")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the
         formula.")
        - nelements: int = Field(None, description="Number of elements.")
        - nsites: int = Field(None, description="Total number of sites in the
          structure.")
        - structure: Structure = Field(None, title = "The Pymatgen Structure
          object from the Atoms object, if
         periodic and store_pmg is True.")
        - symmetry: SymmetryData = Field(None, description="Symmetry data for
          this material.")
            - symmetry.crystal_system: CrystalSystem = Field(None,
              title="Crystal System", description="The crystal system for this
              lattice.")
            - symmetry.number: int = Field(None, title="Space Group Number",
              description="The spacegroup number for the lattice.")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice.")
            - symmetry.symbol: str = Field(None, title="Space Group Symbol",
              description="The spacegroup symbol for the lattice.")
            - symmetry.symprec: float = Field(None, title="Symmetry Finding
              Precision", description="The precision given to spglib to
              determine the symmetry of this lattice.")
        - volume: float = Field(None, title="Volume", description="Total volume
          for this structure in Angstroms^3.")

        For molecules that lack periodicity, the task document also has the
        following fields:

        - charge: int = Field(None, description="Charge of the molecule")
        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in the
         molecule")
        - composition: Composition = Field(None, description="Full composition
          for the molecule")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified
         representation of the composition")
        - elements: List[Element] = Field(None, description="List of elements in
          the molecule")
        - formula_alphabetical: str = Field(None, title="Alphabetical Formula",
          description="Alphabetical molecular
         formula")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation of
         the formula")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the
         formula.")
        - molecule: Molecule = Field(None, title = "The Pymatgen Molecule object
          from the Atoms object, if not
         periodic and store_pmg is True.")
        - natoms: int = Field(None, description="Total number of atoms in the
          molecule")
        - nelectrons: int = Field(None, title="Number of electrons",
          description="The total number of electrons
         for the molecule")
        - nelements: int = Field(None, title="Number of Elements")
        - spin_multiplicity: int = Field(None, description="Spin multiplicity of
          the molecule")
        - symmetry: PointGroupData = Field(None, description="Symmetry data for
          this molecule")
            - symmetry.eigen_tolerance: float = Field(None, title="Interia
              Tensor Eigenvalue Tolerance", description="Tolerance to compare
              eigen values of the inertia tensor.")
            - symmetry.linear: bool = Field(None, title="Molecule Linearity",
              description="Is the molecule linear?")
            - symmetry.matrix_tolerance: float = Field(None, title="Symmetry
              Operation Matrix Element Tolerance" description="Tolerance used to
              generate the full set of symmetry operations of the point group.")
            - symmetry.rotation_number: float = Field(None, title="Rotational
              Symmetry Number", description="Rotational symmetry number for the
              molecule")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice")
            - symmetry.tolerance: float = Field(None, title="Point Group
              Analyzer Tolerance", description="Distance tolerance to consider
              sites as symmetrically equivalent.")
    """
    # Make sure there is a calculator with results
    if not atoms.calc:
        msg = "ASE Atoms object has no attached calculator."
        raise ValueError(msg)
    if not atoms.calc.results:
        msg = "ASE Atoms object's calculator has no results."
        raise ValueError(msg)
    store = SETTINGS.PRIMARY_STORE if store is None else store

    additional_fields = additional_fields or {}

    # Fetch all tabulated results from the attached calculator
    results = {"results": atoms.calc.results}

    # Get the calculator inputs
    uri = get_uri(Path.cwd())
    inputs = {
        "parameters": atoms.calc.parameters,
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
    }
    if input_atoms:
        input_atoms_db = atoms_to_metadata(
            input_atoms, charge_and_multiplicity=charge_and_multiplicity
        )
        inputs["input_atoms"] = input_atoms_db

    # Prepares the Atoms object for the next run by moving the final magmoms to
    # initial, clearing the calculator state, and assigning the resulting Atoms
    # object a unique ID.
    if prep_next_run:
        atoms = prep_next_run_(atoms)

    # Get tabulated properties of the structure itself
    atoms_db = atoms_to_metadata(atoms, charge_and_multiplicity=charge_and_multiplicity)

    # Create a dictionary of the inputs/outputs
    task_doc = clean_dict(
        atoms_db | inputs | results | additional_fields, remove_empties=remove_empties
    )

    if store:
        results_to_db(store, task_doc)

    return task_doc


def summarize_opt_run(
    dyn: Optimizer,
    trajectory: Trajectory | list[Atoms] = None,
    check_convergence: bool | None = None,
    charge_and_multiplicity: tuple[int, int] | None = None,
    prep_next_run: bool = True,
    remove_empties: bool = False,
    additional_fields: dict | None = None,
    store: Store | None = None,
) -> OptSchema:
    """
    Get tabulated results from an ASE Atoms trajectory and store them in a
    database-friendly format. This is meant to be compatible with all calculator
    types.

    Parameters
    ----------
    dyn
        ASE Optimizer object.
    trajectory
        ASE Trajectory object or list[Atoms] from reading a trajectory file. If
        None, the trajectory must be found in dyn.traj_atoms.
    check_convergence
        Whether to check the convergence of the calculation. Defaults to True in
        settings.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    prep_next_run
        Whether the Atoms object stored in {"atoms": atoms} should be prepared
        for the next run This clears out any attached calculator and moves the
        final magmoms to the initial magmoms.
    remove_empties
        Whether to remove None values and empty lists/dicts from the task
        document.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    dict
        Dictionary representation of the task document with the following
        fields:

        - atoms: Atoms = Field(None, title = "The Atoms object from the
          calculation result.")
        - atoms_info: dict = Field(None, title = "The Atoms object info obtained
          from atoms.info.")
        - builder_meta: EmmetMeta = Field(default_factory=EmmetMeta,
          description="Builder metadata."):
            - builder_meta.build_date: str =
              Field(default_factory=datetime.utcnow, description="The build date
              for this document.")
            - builder_meta.database_version: str = Field(None, description="The
              database version for the built data.")
            - builder_meta.emmet_version: str = Field(__version__,
              description="The version of emmet this document was built with.")
            - builder_meta.pymatgen_version: str = Field(pmg_version,
              description="The version of pymatgen this document was built
              with.")
            - builder_meta.pull_request: int = Field(None, description="The pull
              request number associated with this data build.")
        - dir_name: str = Field(None, description="Directory where the output is
          parsed")
        - input_structure: Molecule | Structure = Field(None, title = "The
          Pymatgen Structure or Molecule object from
         the input Atoms object if input_atoms is not None.")
        - nid: str = Field(None, title = "The node ID representing the machine
          where the calculation was run.")
        - parameters: dict = Field(None, title = "the parameters used to run the
          calculation.")
        - opt_parameters: dict = Field(None, title = "the parameters used to run
          the optimization.")
        - results: dict = Field(None, title = "The results from the
          calculation.")
        - trajectory: List[Atoms] = Trajectory of Atoms objects
        - trajectory_results: List[dict] = List of ase.calc.results from the
          trajectory

        For periodic structures, the task document also has the following
        fields:

        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in the material.")
        - composition: Composition = Field(None, description="Full composition
          for the material.")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified
         representation of the composition.")
        - density: float = Field(None, title="Density", description="Density in
          grams per cm^3.")
        - density_atomic: float = Field(None, title="Packing Density",
          description="The atomic packing density in
         atoms per cm^3.")
        - elements: List[Element] = Field(None, description="List of elements in
          the material.")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation
         of the formula.")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the
         formula.")
        - nelements: int = Field(None, description="Number of elements.")
        - nsites: int = Field(None, description="Total number of sites in the
          structure.")
        - structure: Structure = Field(None, title = "The Pymatgen Structure
          object from the Atoms object, if periodic
         and store_pmg is True.")
        - symmetry: SymmetryData = Field(None, description="Symmetry data for
          this material.")
            - symmetry.crystal_system: CrystalSystem = Field(None,
              title="Crystal System", description="The crystal system for this
              lattice.")
            - symmetry.number: int = Field(None, title="Space Group Number",
              description="The spacegroup number for the lattice.")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice.")
            - symmetry.symbol: str = Field(None, title="Space Group Symbol",
              description="The spacegroup symbol for the lattice.")
            - symmetry.symprec: float = Field(None, title="Symmetry Finding
              Precision", description="The precision given to spglib to
              determine the symmetry of this lattice.")
        - volume: float = Field(None, title="Volume", description="Total volume
          for this structure in Angstroms^3.")

        For molecules that lack periodicity, the task document also has the
        following fields:

        - charge: int = Field(None, description="Charge of the molecule")
        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in
         the molecule")
        - composition: Composition = Field(None, description="Full composition
          for the molecule")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified
         representation of the composition")
        - elements: List[Element] = Field(None, description="List of elements in
          the molecule")
        - formula_alphabetical: str = Field(None, title="Alphabetical Formula",
          description="Alphabetical molecular
         formula")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation
         of the formula")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the
         formula.")
        - molecule: Molecule = Field(None, title = "The Pymatgen Molecule object
          from the Atoms object, if not
         periodic and store_pmg is True.")
        - natoms: int = Field(None, description="Total number of atoms in the
          molecule")
        - nelectrons: int = Field(None, title="Number of electrons",
          description="The total number of electrons for
         the molecule")
        - nelements: int = Field(None, title="Number of Elements")
        - spin_multiplicity: int = Field(None, description="Spin multiplicity of
          the molecule")
        - symmetry: PointGroupData = Field(None, description="Symmetry data for
          this molecule")
            - symmetry.eigen_tolerance: float = Field(None, title="Interia
              Tensor Eigenvalue Tolerance", description="Tolerance to compare
              eigen values of the inertia tensor.")
            - symmetry.linear: bool = Field(None, title="Molecule Linearity",
              description="Is the molecule linear?")
            - symmetry.matrix_tolerance: float = Field(None, title="Symmetry
              Operation Matrix Element Tolerance" description="Tolerance used to
              generate the full set of symmetry operations of the point group.")
            - symmetry.rotation_number: float = Field(None, title="Rotational
              Symmetry Number", description="Rotational symmetry number for the
              molecule")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice")
            - symmetry.tolerance: float = Field(None, title="Point Group
              Analyzer Tolerance", description="Distance tolerance to consider
              sites as symmetrically equivalent.")
    """

    check_convergence = (
        SETTINGS.CHECK_CONVERGENCE if check_convergence is None else check_convergence
    )
    additional_fields = additional_fields or {}
    opt_parameters = dyn.todict()
    if hasattr(dyn, "fmax"):
        opt_parameters = opt_parameters | {"fmax": dyn.fmax}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    # Check convergence
    is_converged = dyn.converged()
    if check_convergence and not is_converged:
        msg = "Optimization did not converge."
        raise ValueError(msg)

    # Get trajectory
    if not trajectory:
        trajectory = (
            dyn.traj_atoms
            if hasattr(dyn, "traj_atoms")
            else read(dyn.trajectory.filename, index=":")
        )

    initial_atoms = trajectory[0]
    final_atoms = dyn.atoms.atoms if isinstance(dyn.atoms, Filter) else dyn.atoms

    # Get results
    traj_results = {
        "trajectory_results": [atoms.calc.results for atoms in trajectory],
        "trajectory": [
            atoms_to_metadata(atoms, charge_and_multiplicity=charge_and_multiplicity)
            for atoms in trajectory
        ],
    }

    results = {
        "results": final_atoms.calc.results
        | {"converged": is_converged, "nsteps": dyn.get_number_of_steps()}
    }

    # Get the calculator inputs
    uri = get_uri(Path.cwd())
    inputs = {
        "parameters": dyn.atoms.calc.parameters,
        "parameters_opt": opt_parameters,
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
    }
    input_atoms_db = atoms_to_metadata(
        initial_atoms, charge_and_multiplicity=charge_and_multiplicity
    )
    inputs["input_structure"] = input_atoms_db

    # Prepares the Atoms object for the next run by moving the final magmoms to
    # initial, clearing the calculator state, and assigning the resulting Atoms
    # object a unique ID.
    if prep_next_run:
        final_atoms = prep_next_run_(final_atoms)

    # Get tabulated properties of the structure itself
    atoms_db = atoms_to_metadata(
        final_atoms, charge_and_multiplicity=charge_and_multiplicity
    )

    # Create a dictionary of the inputs/outputs
    task_doc = clean_dict(
        atoms_db | inputs | results | traj_results | additional_fields,
        remove_empties=remove_empties,
    )

    if store:
        results_to_db(store, task_doc)

    return task_doc


def summarize_vib_run(
    vib: Vibrations,
    charge_and_multiplicity: tuple[int, int] | None = None,
    remove_empties: bool = False,
    additional_fields: dict | None = None,
    store: Store | None = None,
) -> VibSchema:
    """
    Get tabulated results from an ASE Vibrations object and store them in a
    database-friendly format.

    Parameters
    ----------
    vib
        ASE Vibrations object.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    remove_empties
        Whether to remove None values and empty lists/dicts from the task
        document.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    dict
        Dictionary representation of the task document with the following
        fields:

        - atoms: Atoms = Field(None, title = "The Atoms object from the
          calculation result.")
        - atoms_info: dict = Field(None, title = "The Atoms object info obtained
          from atoms.info.")
        - builder_meta: EmmetMeta = Field(default_factory=EmmetMeta,
          description="Builder metadata."):
            - builder_meta.build_date: str =
              Field(default_factory=datetime.utcnow, description="The build date
              for this document.")
            - builder_meta.database_version: str = Field(None, description="The
              database version for the built data.")
            - builder_meta.emmet_version: str = Field(__version__,
              description="The version of emmet this document was built with.")
            - builder_meta.pymatgen_version: str = Field(pmg_version,
              description="The version of pymatgen this document was built
              with.")
            - builder_meta.pull_request: int = Field(None, description="The pull
              request number associated with this data build.")
        - dir_name: str = Field(None, description="Directory where the output is
          parsed")
        - nid: str = Field(None, title = "The node ID representing the machine
          where the calculation was run.")
        - parameters: dict = Field(None, title = "the parameters used to run the
          calculation.")
        - vib_parameters: dict = Field(None, title = "the parameters used to run
          the vibrations.")
            - vib_parameters.delta: float = the Vibrations delta value
            - vib_parameters.direction: str = the Vibrations direction value
            - vib_parameters.method: str = the Vibrations method value
            - vib_parameters.ndof: int = the Vibrations ndof value
            - vib_parameters.nfree: int = the Vibrations nfree value
        - results: dict = Field(None, title = "The results from the
          calculation.")
            - results.imag_vib_freqs: List[float] = Imaginary vibrational
              frequencies in cm^-1
            - results.n_imag: int = Number of imaginary vibrational frequencies
            - results.vib_energies: List[float] = Vibrational energies in eV.
              3N-5 or 3N-6 for molecules; 3N for solids.
            - results.vib_freqs: List[float] = Vibrational frequencies in cm^-1.
              3N-5 or 3N-6 for molecules; 3N for solids.
            - results.vib_energies_raw: List[float] = Vibrational energies in eV
              of length 3N.
            - results.vib_freqs_raw: List[float] = Vibrational frequencies in
              cm^-1 of length 3N.

        For periodic structures, the task document also has the following
        fields:

        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in the material.")
        - composition: Composition = Field(None, description="Full composition
          for the material.")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified
         representation of the composition.")
        - density: float = Field(None, title="Density", description="Density in
          grams per cm^3.")
        - density_atomic: float = Field(None, title="Packing Density",
          description="The atomic packing density in atoms
         per cm^3.")
        - elements: List[Element] = Field(None, description="List of elements in
          the material.")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation of the
         formula.")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the
         formula.")
        - nelements: int = Field(None, description="Number of elements.")
        - nsites: int = Field(None, description="Total number of sites in the
          structure.")
        - structure: Structure = Field(None, title = "The Pymatgen Structure
          object from the Atoms object, if periodic
         and store_pmg is True.")
        - symmetry: SymmetryData = Field(None, description="Symmetry data for
          this material.")
            - symmetry.crystal_system: CrystalSystem = Field(None,
              title="Crystal System", description="The crystal system for this
              lattice.")
            - symmetry.number: int = Field(None, title="Space Group Number",
              description="The spacegroup number for the lattice.")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice.")
            - symmetry.symbol: str = Field(None, title="Space Group Symbol",
              description="The spacegroup symbol for the lattice.")
            - symmetry.symprec: float = Field(None, title="Symmetry Finding
              Precision", description="The precision given to spglib to
              determine the symmetry of this lattice.")
        - volume: float = Field(None, title="Volume", description="Total volume
          for this structure in Angstroms^3.")

        For molecules that lack periodicity, the task document also has the
        following fields:

        - charge: int = Field(None, description="Charge of the molecule")
        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in the
         molecule")
        - composition: Composition = Field(None, description="Full composition
          for the molecule")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified
         representation of the composition")
        - elements: List[Element] = Field(None, description="List of elements in
          the molecule")
        - formula_alphabetical: str = Field(None, title="Alphabetical Formula",
          description="Alphabetical molecular
         formula")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation of
         the formula")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the
         formula.")
        - natoms: int = Field(None, description="Total number of atoms in the
          molecule")
        - molecule: Molecule = Field(None, title = "The Pymatgen Molecule object
          from the Atoms object, if not
         periodic and store_pmg is True.")
        - nelectrons: int = Field(None, title="Number of electrons",
          description="The total number of electrons for
         the molecule")
        - nelements: int = Field(None, title="Number of Elements")
        - spin_multiplicity: int = Field(None, description="Spin multiplicity of
          the molecule")
        - symmetry: PointGroupData = Field(None, description="Symmetry data for
          this molecule")
            - symmetry.eigen_tolerance: float = Field(None, title="Interia
              Tensor Eigenvalue Tolerance", description="Tolerance to compare
              eigen values of the inertia tensor.")
            - symmetry.linear: bool = Field(None, title="Molecule Linearity",
              description="Is the molecule linear?")
            - symmetry.matrix_tolerance: float = Field(None, title="Symmetry
              Operation Matrix Element Tolerance" description="Tolerance used to
              generate the full set of symmetry operations of the point group.")
            - symmetry.rotation_number: float = Field(None, title="Rotational
              Symmetry Number", description="Rotational symmetry number for the
              molecule")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice")
            - symmetry.tolerance: float = Field(None, title="Point Group
              Analyzer Tolerance", description="Distance tolerance to consider
              sites as symmetrically equivalent.")
    """
    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    vib_freqs_raw = vib.get_frequencies().tolist()
    vib_energies_raw = vib.get_energies().tolist()

    atoms = vib._atoms if isinstance(vib, VibrationsData) else vib.atoms

    # Convert imaginary modes to negative values for DB storage
    for i, f in enumerate(vib_freqs_raw):
        if np.imag(f) > 0:
            vib_freqs_raw[i] = -np.abs(f)
            vib_energies_raw[i] = -np.abs(vib_energies_raw[i])
        else:
            vib_freqs_raw[i] = np.abs(f)
            vib_energies_raw[i] = np.abs(vib_energies_raw[i])

    uri = get_uri(Path.cwd())
    inputs = {
        "parameters": None
        if isinstance(vib, VibrationsData)
        else atoms.calc.parameters,
        "parameters_vib": None
        if isinstance(vib, VibrationsData)
        else {
            "delta": vib.delta,
            "direction": vib.direction,
            "method": vib.method,
            "ndof": vib.ndof,
            "nfree": vib.nfree,
        },
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
    }

    atoms_db = atoms_to_metadata(atoms, charge_and_multiplicity=charge_and_multiplicity)

    # Get the true vibrational modes
    natoms = len(atoms)
    if natoms == 1:
        vib_freqs = []
        vib_energies = []
    elif atoms.pbc.any():
        vib_freqs = vib_freqs_raw
        vib_energies = vib_energies_raw
    else:
        # Sort by absolute value
        vib_freqs_raw_sorted = vib_freqs_raw.copy()
        vib_energies_raw_sorted = vib_energies_raw.copy()
        vib_freqs_raw_sorted.sort(key=np.abs)
        vib_energies_raw_sorted.sort(key=np.abs)

        # Cut the 3N-5 or 3N-6 modes based on their absolute value
        n_modes = 3 * natoms - 5 if atoms_db["symmetry"]["linear"] else 3 * natoms - 6
        vib_freqs = vib_freqs_raw_sorted[-n_modes:]
        vib_energies = vib_energies_raw_sorted[-n_modes:]

    imag_vib_freqs = [f for f in vib_freqs if f < 0]

    results = {
        "results": {
            "imag_vib_freqs": imag_vib_freqs,
            "n_imag": len(imag_vib_freqs),
            "vib_energies": vib_energies,
            "vib_freqs": vib_freqs,
            "vib_energies_raw": vib_energies_raw,
            "vib_freqs_raw": vib_freqs_raw,
        }
    }

    task_doc = clean_dict(
        atoms_db | inputs | results | additional_fields, remove_empties=remove_empties
    )

    if store:
        results_to_db(store, task_doc)

    return task_doc


def summarize_thermo(
    igt: IdealGasThermo,
    temperature: float = 298.15,
    pressure: float = 1.0,
    charge_and_multiplicity: tuple[int, int] | None = None,
    remove_empties: bool = False,
    additional_fields: dict | None = None,
    store: Store | None = None,
) -> ThermoSchema:
    """
    Get tabulated results from an ASE IdealGasThermo object and store them in a
    database-friendly format.

    Parameters
    ----------
    igt
        ASE IdealGasThermo object.
    temperature
        Temperature in Kelvins.
    pressure
        Pressure in bar.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    remove_empties
        Whether to remove None values and empty lists/dicts from the task
        document.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    dict
        Dictionary representation of the task document with the following
        fields:

        - atoms: Atoms = Field(None, title = "The Atoms object from the
          calculation result.")
        - atoms_info: dict = Field(None, title = "The Atoms object info obtained
          from atoms.info.")
        - builder_meta: EmmetMeta = Field(default_factory=EmmetMeta,
          description="Builder metadata."):
            - builder_meta.build_date: str =
              Field(default_factory=datetime.utcnow, description="The build date
              for this document.")
            - builder_meta.database_version: str = Field(None, description="The
              database version for the built data.")
            - builder_meta.emmet_version: str = Field(__version__,
              description="The version of emmet this document was built with.")
            - builder_meta.pymatgen_version: str = Field(pmg_version,
              description="The version of pymatgen this document was built
              with.")
            - builder_meta.pull_request: int = Field(None, description="The pull
              request number associated with this data build.")
        - dir_name: str = Field(None, description="Directory where the output is
          parsed")
        - nid: str = Field(None, title = "The node ID representing the machine
          where the calculation was run.")
        - thermo_parameters: dict = Field(None, title = "the parameters used to
          run the thermo calculation.")
            - thermo_parameters.temperature: float = Temperature in Kelvins
            - thermo_parameters.pressure: float = Pressure in bar
            - thermo_parameters.sigma: float = The rotational symmetry number of
              the molecule
            - thermo_parameters.spin_multiplicity: int = The spin multiplicity
              of the molecule
            - thermo_parameters.vib_freqs: List[float] = Vibrational frequencies
              in cm^-1 used for the thermo calculation
            - thermo_parameters.vib_energies: List[float] = Vibrational energies
              in eV used for the thermo calculation
            - thermo_parameters.n_imag: int = Number of imaginary vibrational
              frequencies ignored in the thermo calculation
        - results: dict = Field(None, title = "The results from the
          calculation.")
            - results.energy: float = The potential energy of the system in eV
            - results.enthalpy: float = The enthalpy of the system in eV
            - results.entropy: float = The entropy of the system in eV/K
            - results.gibbs_energy: float = The Gibbs free energy of the system
              in eV
            - results.zpe: float = The zero point vibrational energy of the
              system in eV

        The task document also has the following fields from the Molecule
        object:

        - charge: int = Field(None, description="Charge of the molecule")
        - chemsys: str = Field(None, title="Chemical System",
          description="dash-delimited string of elements in the
         molecule")
        - composition: Composition = Field(None, description="Full composition
          for the molecule")
        - composition_reduced: Composition = Field(None, title="Reduced
          Composition", description="Simplified
         representation of the composition")
        - elements: List[Element] = Field(None, description="List of elements in
          the molecule")
        - formula_alphabetical: str = Field(None, title="Alphabetical Formula",
          description="Alphabetical molecular
         formula")
        - formula_anonymous: str = Field(None, title="Anonymous Formula",
          description="Anonymized representation of
         the formula")
        - formula_pretty: str = Field(None, title="Pretty Formula",
          description="Cleaned representation of the
         formula.")
        - natoms: int = Field(None, description="Total number of atoms in the
          molecule")
        - molecule: Molecule = Field(None, title = "The Pymatgen Molecule object
          from the Atoms object, if not
         periodic and store_pmg is True.")
        - nelectrons: int = Field(None, title="Number of electrons",
          description="The total number of electrons for
         the molecule")
        - nelements: int = Field(None, title="Number of Elements")
        - spin_multiplicity: int = Field(None, description="Spin multiplicity of
          the molecule")
        - symmetry: PointGroupData = Field(None, description="Symmetry data for
          this molecule")
            - symmetry.eigen_tolerance: float = Field(None, title="Interia
              Tensor Eigenvalue Tolerance", description="Tolerance to compare
              eigen values of the inertia tensor.")
            - symmetry.linear: bool = Field(None, title="Molecule Linearity",
              description="Is the molecule linear?")
            - symmetry.matrix_tolerance: float = Field(None, title="Symmetry
              Operation Matrix Element Tolerance" description="Tolerance used to
              generate the full set of symmetry operations of the point group.")
            - symmetry.rotation_number: float = Field(None, title="Rotational
              Symmetry Number", description="Rotational symmetry number for the
              molecule")
            - symmetry.point_group: str = Field(None, title="Point Group
              Symbol", description="The point group for the lattice")
            - symmetry.tolerance: float = Field(None, title="Point Group
              Analyzer Tolerance", description="Distance tolerance to consider
              sites as symmetrically equivalent.")
    """

    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    uri = get_uri(Path.cwd())
    spin_multiplicity = int(2 * igt.spin + 1)

    inputs = {
        "parameters_thermo": {
            "temperature": temperature,
            "pressure": pressure,
            "sigma": igt.sigma,
            "spin_multiplicity": spin_multiplicity,
            "vib_freqs": [e / units.invcm for e in igt.vib_energies],
            "vib_energies": igt.vib_energies.tolist(),
            "n_imag": igt.n_imag,
        },
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
    }

    results = {
        "results": {
            "energy": igt.potentialenergy,
            "enthalpy": igt.get_enthalpy(temperature, verbose=True),
            "entropy": igt.get_entropy(temperature, pressure * 10**5, verbose=True),
            "gibbs_energy": igt.get_gibbs_energy(
                temperature, pressure * 10**5, verbose=True
            ),
            "zpe": igt.get_ZPE_correction(),
        }
    }

    if charge_and_multiplicity and spin_multiplicity != charge_and_multiplicity[1]:
        warnings.warn(
            "The IdealGasThermo spin multiplicity does not match the user-specified multiplicity.",
            UserWarning,
        )

    atoms_db = atoms_to_metadata(
        igt.atoms, charge_and_multiplicity=charge_and_multiplicity
    )

    task_doc = clean_dict(
        atoms_db | inputs | results | additional_fields, remove_empties=remove_empties
    )

    if store:
        results_to_db(store, task_doc)

    return task_doc



"""Transition state recipes for the Q-Chem"""
from __future__ import annotations

import multiprocessing
from typing import TYPE_CHECKING

from monty.dev import requires

from quacc import SETTINGS, job
from quacc.calculators.qchem import QChem
from quacc.recipes.qchem.core import relax_job
from quacc.schemas import fetch_atoms
from quacc.schemas.ase import summarize_opt_run
from quacc.utils.atoms import check_charge_and_spin
from quacc.utils.calc import run_ase_opt
from quacc.utils.dicts import merge_dicts, remove_dict_empties

try:
    from sella import IRC, Sella

    has_sella = True
except ImportError:
    has_sella = False

if TYPE_CHECKING:
    from typing import Literal

    from ase import Atoms

    from quacc.schemas.ase import OptSchema


@job
@requires(
    has_sella,
    "Sella must be installed. Try pip install quacc[optimizers].",
)
def ts_job(
    atoms: Atoms | dict,
    charge: int | None = None,
    spin_multiplicity: int | None = None,
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict | None = None,
    opt_swaps: dict | None = None,
) -> OptSchema:
    """
    TS optimize a molecular structure.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    charge
        The total charge of the molecular system. Effectively defaults to zero.
    spin_multiplicity
        The spin multiplicity of the molecular system. Effectively defaults to
        the lowest spin state given the molecular structure and charge.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-SVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum impicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to pymatgen.io.qchem.QChemDictSet which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    opt_swaps
        Dictionary of custom kwargs for `run_ase_opt`. Overrides the following
        defaults:

        ```python
        {"fmax": 0.01, "max_steps": 1000, "optimizer": "Sella"}
        ```

    Returns
    -------
    OptSchema
        Dictionary of results from `quacc.schemas.ase.summarize_opt_run`
    """

    # TODO:
    #   - exposing TRICs?
    #   - passing initial Hessian?
    atoms = fetch_atoms(atoms)
    checked_charge, checked_spin_multiplicity = check_charge_and_spin(
        atoms, charge, spin_multiplicity
    )

    qchem_defaults = {
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or multiprocessing.cpu_count(),
        "qchem_input_params": {
            "basis_set": basis,
            "scf_algorithm": scf_algorithm,
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else None,
        },
    }
    qchem_flags = remove_dict_empties(qchem_defaults)

    opt_swaps = opt_swaps or {}
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": Sella,
    }
    opt_flags = merge_dicts(opt_defaults, opt_swaps)
    if opt_flags["optimizer"] != Sella:
        raise ValueError("Only Sella should be used for TS optimization.")

    atoms.calc = QChem(atoms, **qchem_flags)
    dyn = run_ase_opt(atoms, **opt_flags)

    return summarize_opt_run(
        dyn,
        charge_and_multiplicity=(checked_charge, checked_spin_multiplicity),
        additional_fields={"name": "Q-Chem TS Optimization"},
    )


@job
@requires(
    has_sella,
    "Sella must be installed. Try pip install quacc[optimizers].",
)
def irc_job(
    atoms: Atoms | dict,
    direction: Literal["forward", "reverse"],
    charge: int | None = None,
    spin_multiplicity: int | None = None,
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict | None = None,
    opt_swaps: dict | None = None,
) -> OptSchema:
    """
    IRC optimize a molecular structure.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    direction
        Direction of the IRC. Should be "forward" or "reverse".
    charge
        The total charge of the molecular system. Effectively defaults to zero.
    spin_multiplicity
        The spin multiplicity of the molecular system. Effectively defaults to
        the lowest spin state given the molecular structure and charge.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-SVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum impicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to pymatgen.io.qchem.QChemDictSet which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    opt_swaps
        Dictionary of custom kwargs for `run_ase_opt`. Overrides the following
        defaults:

        ```python
        {"fmax": 0.01, "max_steps": 1000, "optimizer": "Sella"}
        ```

    Returns
    -------
    OptSchema
        Dictionary of results from `quacc.schemas.ase.summarize_opt_run`
    """

    # TODO: 1) expose TRICs?; 2) passing initial Hessian?
    atoms = fetch_atoms(atoms)
    checked_charge, checked_spin_multiplicity = check_charge_and_spin(
        atoms, charge, spin_multiplicity
    )

    qchem_defaults = {
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or multiprocessing.cpu_count(),
        "qchem_input_params": {
            "basis_set": basis,
            "scf_algorithm": scf_algorithm,
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else None,
        },
    }
    qchem_flags = remove_dict_empties(qchem_defaults)

    opt_swaps = opt_swaps or {}
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": IRC,
        "optimizer_kwargs": {"keep_going": True},
        "run_kwargs": {"direction": direction},
    }
    opt_flags = merge_dicts(opt_defaults, opt_swaps)
    if opt_flags["optimizer"] != IRC:
        raise ValueError("Only Sella's IRC should be used for IRC optimization.")

    atoms.calc = QChem(atoms, **qchem_flags)
    dyn = run_ase_opt(atoms, **opt_flags)

    return summarize_opt_run(
        dyn,
        charge_and_multiplicity=(checked_charge, checked_spin_multiplicity),
        additional_fields={"name": "Q-Chem IRC Optimization"},
    )


@job
@requires(
    has_sella,
    "Sella must be installed. Try pip install quacc[optimizers].",
)
def quasi_irc_job(
    atoms: Atoms | dict,
    direction: Literal["forward", "reverse"],
    shared_kwargs: dict | None = None,
    irc_opt_swaps: dict | None = None,
    relax_opt_swaps: dict | None = None,
) -> OptSchema:
    """
    Quasi-IRC optimize a molecular structure.

    Parameters
    ----------
    atoms
        Atoms object.
    direction
        Direction of the IRC. Should be "forward" or "reverse".
    shared_kwargs
        Dictionary of kwargs that are passed as input to both irc_job and
        relax_job.
    irc_opt_swaps
        Dictionary of opt_swap kwargs for the irc_job. Overrides the following
        defaults:

        ```python
        {"fmax": 100, "max_steps": 10}
        ```
    relax_opt_swaps
        Dictionary of opt_swap kwargs for the relax_job.

    Returns
    -------
    OptSchema
        Dictionary of results from `quacc.schemas.ase.summarize_opt_run`
    """

    shared_kwargs = shared_kwargs or {}
    irc_opt_swaps = irc_opt_swaps or {}
    relax_opt_swaps = relax_opt_swaps or {}
    default_settings = SETTINGS.copy()

    irc_opt_swaps_defaults = {
        "fmax": 100,
        "max_steps": 10,
    }
    irc_opt_swaps = merge_dicts(irc_opt_swaps_defaults, irc_opt_swaps)

    SETTINGS.CHECK_CONVERGENCE = False
    irc_summary = irc_job.__wrapped__(
        atoms,
        direction=direction,
        opt_swaps=irc_opt_swaps,
        **shared_kwargs,
    )

    SETTINGS.CHECK_CONVERGENCE = default_settings.CHECK_CONVERGENCE
    relax_summary = relax_job(
        irc_summary,
        opt_swaps=relax_opt_swaps,
        **shared_kwargs,
    )

    relax_summary["initial_irc"] = irc_summary

    return relax_summary



"""Core recipes for ORCA"""
from __future__ import annotations

import multiprocessing
from shutil import which
from typing import TYPE_CHECKING

from ase.calculators.orca import ORCA, OrcaProfile

from quacc import SETTINGS, job
from quacc.schemas import fetch_atoms
from quacc.schemas.cclib import summarize_run
from quacc.utils.atoms import get_charge, get_multiplicity
from quacc.utils.calc import run_calc
from quacc.utils.dicts import merge_dicts

if TYPE_CHECKING:
    from ase import Atoms

    from quacc.schemas.cclib import cclibSchema


LOG_FILE = f"{ORCA().name}.out"
GEOM_FILE = f"{ORCA().name}.xyz"


@job
def static_job(
    atoms: Atoms | dict,
    charge: int | None = None,
    multiplicity: int | None = None,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    input_swaps: dict | None = None,
    block_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> cclibSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object or a dictionary with the key "atoms" and an Atoms object as
        the value
    charge
        Charge of the system. If None, this is determined from the sum of
        `atoms.get_initial_charges()`.
    multiplicity
        Multiplicity of the system. If None, this is determined from 1+ the sum
        of `atoms.get_initial_magnetic_moments()`.
    xc
        Exchange-correlation functional
    basis
        Basis set
    input_swaps
        Dictionary of orcasimpleinput swaps for the calculator. To enable new
        entries, set the value as True. To remove entries from the defaults, set
        the value as None. Overrides the following defaults:

        ```python
        {
            xc: True,
            basis: True,
            "sp": True,
            "slowconv": True,
            "normalprint": True,
            "xyzfile": True,
        }
        ```
    block_swaps
        Dictionary of orcablock swaps for the calculator. To enable new entries,
        set the value as True. To remove entries from the defaults, set the
        value as None. Overrides the following defaults:

        ```python
        (
            {f"%pal nprocs {multiprocessing.cpu_count()} end": True}
            if which("mpirun")
            else {}
        )
        ```

    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    cclibSchema
        Dictionary of results from `quacc.schemas.cclib.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    input_swaps = input_swaps or {}
    block_swaps = block_swaps or {}

    default_inputs = {
        xc: True,
        basis: True,
        "sp": True,
        "slowconv": True,
        "normalprint": True,
        "xyzfile": True,
    }
    default_blocks = (
        {f"%pal nprocs {multiprocessing.cpu_count()} end": True}
        if which("mpirun")
        else {}
    )

    inputs = merge_dicts(default_inputs, input_swaps)
    blocks = merge_dicts(default_blocks, block_swaps)
    orcasimpleinput = " ".join(list(inputs.keys()))
    orcablocks = " ".join(list(blocks.keys()))

    atoms.calc = ORCA(
        profile=OrcaProfile([SETTINGS.ORCA_CMD]),
        charge=get_charge(atoms) if charge is None else charge,
        mult=get_multiplicity(atoms) if multiplicity is None else multiplicity,
        orcasimpleinput=orcasimpleinput,
        orcablocks=orcablocks,
    )
    atoms = run_calc(atoms, geom_file=GEOM_FILE, copy_files=copy_files)

    return summarize_run(
        atoms,
        LOG_FILE,
        additional_fields={"name": "ORCA Static"},
    )


@job
def relax_job(
    atoms: Atoms | dict,
    charge: int | None = None,
    multiplicity: int | None = None,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    run_freq: bool = False,
    input_swaps: dict | None = None,
    block_swaps: dict | None = None,
    copy_files: list[str] | None = None,
) -> cclibSchema:
    """
    Carry out a geometry optimization.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system. If None, this is determined from the sum of
        atoms.get_initial_charges().
    multiplicity
        Multiplicity of the system. If None, this is determined from 1+ the sum
        of atoms.get_initial_magnetic_moments().
    xc
        Exchange-correlation functional
    basis
        Basis set
    run_freq
        If a requency calculation should be carried out.
    input_swaps
        Dictionary of orcasimpleinput swaps for the calculator. To enable new
        entries, set the value as True. To remove entries from the defaults, set
        the value as None. Overrides the following defaults:

        ```python
        {
            xc: True,
            basis: True,
            "opt": True,
            "slowconv": True,
            "normalprint": True,
            "freq": True if run_freq else None,
            "xyzfile": True,
        }
        ```
    block_swaps
        Dictionary of orcablock swaps for the calculator. To enable new entries,
        set the value as True. To remove entries from the defaults, set the
        value as None. Overrides the following defaults:

        ```python
        (
            {f"%pal nprocs {multiprocessing.cpu_count()} end": True}
            if which("mpirun")
            else {}
        )
        ```
    copy_files
        Files to copy to the runtime directory.

    Returns
    -------
    cclibSchema
        Dictionary of results from `quacc.schemas.cclib.summarize_run`
    """
    atoms = fetch_atoms(atoms)
    input_swaps = input_swaps or {}
    block_swaps = block_swaps or {}

    default_inputs = {
        xc: True,
        basis: True,
        "opt": True,
        "slowconv": True,
        "normalprint": True,
        "freq": True if run_freq else None,
        "xyzfile": True,
    }
    default_blocks = (
        {f"%pal nprocs {multiprocessing.cpu_count()} end": True}
        if which("mpirun")
        else {}
    )

    inputs = merge_dicts(default_inputs, input_swaps)
    blocks = merge_dicts(default_blocks, block_swaps)
    orcasimpleinput = " ".join(list(inputs.keys()))
    orcablocks = " ".join(list(blocks.keys()))

    atoms.calc = ORCA(
        profile=OrcaProfile([SETTINGS.ORCA_CMD]),
        charge=get_charge(atoms) if charge is None else charge,
        mult=get_multiplicity(atoms) if multiplicity is None else multiplicity,
        orcasimpleinput=orcasimpleinput,
        orcablocks=orcablocks,
    )
    atoms = run_calc(atoms, geom_file=GEOM_FILE, copy_files=copy_files)

    return summarize_run(
        atoms,
        LOG_FILE,
        additional_fields={"name": "ORCA Relax"},
    )



"""
Utility functions for dealing with dictionaries
"""
from __future__ import annotations


def merge_dicts(dict1: dict, dict2: dict, remove_empties: bool = True) -> dict:
    """
    Recursively merges two dictionaries.

    Parameters
    ----------

    dict1
        First dictionary
    dict2
        Second dictionary
    remove_empties
        If True, remove empty lists and dictionaries

    Returns
    -------
    dict
        Merged dictionary
    """
    merged = dict1.copy()

    for key, value in dict2.items():
        if key in merged:
            if isinstance(merged[key], dict) and isinstance(value, dict):
                merged[key] = merge_dicts(merged[key], value)
            else:
                merged[key] = value
        else:
            merged[key] = value

    if remove_empties:
        merged = remove_dict_empties(merged)

    return merged


def clean_dict(start_dict: dict, remove_empties: bool = False) -> dict:
    """
    For a given dictionary, recursively remove all items that are None or are
    empty lists/dicts, and then sort all entries alphabetically by key.

    Parameters
    ----------
    start_dict
        Dictionary to clean
    remove_empties
        If True, remove empty lists and dictionaries

    Returns
    -------
    dict
        Cleaned and sorted dictionary
    """

    if remove_empties:
        start_dict = remove_dict_empties(start_dict)
    return sort_dict(start_dict)


def remove_dict_empties(start_dict: dict) -> dict:
    """
    For a given dictionary, recursively remove all items that are None or are
    empty lists/dicts.

    Parameters
    ----------
    start_dict
        Dictionary to clean

    Returns
    -------
    dict
        Cleaned dictionary
    """

    if isinstance(start_dict, dict):
        return {
            k: remove_dict_empties(v)
            for k, v in start_dict.items()
            if v is not None and (not isinstance(v, (dict, list)) or len(v) != 0)
        }
    return (
        [remove_dict_empties(v) for v in start_dict]
        if isinstance(start_dict, list)
        else start_dict
    )


def sort_dict(start_dict: dict) -> dict:
    """
    For a given dictionary, recursively sort all entries alphabetically by key.

    Parameters
    ----------
    start_dict
        Dictionary to sort

    Returns
    -------
    dict
        Sorted dictionary
    """

    return {
        k: sort_dict(v) if isinstance(v, dict) else v
        for k, v in sorted(start_dict.items())
    }



"""A Q-Chem calculator built on Pymatgen and Custodian functionality"""
from __future__ import annotations

import inspect
import struct
from pathlib import Path

from ase import Atoms, units
from ase.calculators.calculator import FileIOCalculator
from monty.io import zopen
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.qchem.outputs import QCOutput
from pymatgen.io.qchem.sets import ForceSet

from quacc.custodian import qchem as custodian_qchem
from quacc.utils.atoms import check_charge_and_spin


class QChem(FileIOCalculator):
    """

    Parameters
    ----------
    atoms
        The Atoms object to be used for the calculation.
    cores
        Number of cores to use for the Q-Chem calculation.
    charge
        The total charge of the molecular system. Effectively defaults to zero.
    spin_multiplicity
        The spin multiplicity of the molecular system. Effectively defaults to
        the lowest spin state given the molecular structure and charge.
    qchem_input_params
        Dictionary of Q-Chem input parameters to be passed to
        pymatgen.io.qchem.sets.ForceSet.
    **fileiocalculator_kwargs
        Additional arguments to be passed to
        ase.calculators.calculator.FileIOCalculator.

    Returns
    -------
    Atoms
        The ASE Atoms object with attached Q-Chem calculator.
    """

    implemented_properties = ["energy", "forces"]

    def __init__(
        self,
        atoms: Atoms,
        charge: None | int = None,
        spin_multiplicity: None | int = None,
        method: str | None = None,
        cores: int = 1,
        qchem_input_params: dict | None = None,
        **fileiocalculator_kwargs,
    ):
        # Assign variables to self
        self.atoms = atoms
        self.cores = cores
        self.charge = charge
        self.spin_multiplicity = spin_multiplicity
        self.qchem_input_params = qchem_input_params or {}
        self.fileiocalculator_kwargs = fileiocalculator_kwargs

        # Sanity checks
        if "directory" in self.fileiocalculator_kwargs:
            raise NotImplementedError("The directory kwarg is not supported.")

        if "overwrite_inputs" not in self.qchem_input_params:
            self.qchem_input_params["overwrite_inputs"] = {}

        if self.charge is None and self.spin_multiplicity is not None:
            raise ValueError("If setting spin_multiplicity, must also specify charge.")

        if self.qchem_input_params.get("smd_solvent") and self.qchem_input_params.get(
            "pcm_dielectric"
        ):
            raise ValueError("PCM and SMD cannot be employed simultaneously.")

        if "rem" not in self.qchem_input_params["overwrite_inputs"]:
            self.qchem_input_params["overwrite_inputs"]["rem"] = {}
        if (
            method
            and "method" not in self.qchem_input_params["overwrite_inputs"]["rem"]
        ):
            self.qchem_input_params["overwrite_inputs"]["rem"]["method"] = method

        # We will save the parameters that have been passed to the Q-Chem
        # calculator via FileIOCalculator's self.default_parameters
        self.default_parameters = {
            "cores": self.cores,
            "charge": self.charge,
            "spin_multiplicity": self.spin_multiplicity,
        }

        # We also want to save the contents of self.qchem_input_params. However,
        # the overwrite_inputs key will have a corresponding value which is
        # either an empty dictionary or a nested dict of dicts, requiring a bit
        # of careful unwrapping.
        for key in self.qchem_input_params:
            if key == "overwrite_inputs":
                for subkey in self.qchem_input_params[key]:
                    for subsubkey in self.qchem_input_params[key][subkey]:
                        self.default_parameters[
                            f"overwrite_{subkey}_{subsubkey}"
                        ] = self.qchem_input_params[key][subkey][subsubkey]
            else:
                self.default_parameters[key] = self.qchem_input_params[key]

        charge, spin_multiplicity = check_charge_and_spin(
            atoms, self.charge, self.spin_multiplicity
        )
        self.charge = charge
        self.spin_multiplicity = spin_multiplicity

        # Get Q-Chem executable command
        self.command = self._manage_environment()

        # Instantiate previous orbital coefficients
        self.prev_orbital_coeffs = None

        # Instantiate the calculator
        FileIOCalculator.__init__(
            self,
            restart=None,
            ignore_bad_restart_file=FileIOCalculator._deprecated,
            label=None,
            atoms=self.atoms,
            **self.fileiocalculator_kwargs,
        )

    def _manage_environment(self) -> str:
        """
        Manage the environment for the Q-Chem calculator.

        Returns
        -------
        str
            The command flag to pass to the Q-Chem calculator.
        """

        # Return the command flag
        run_qchem_custodian_file = Path(inspect.getfile(custodian_qchem)).resolve()
        return f"python {run_qchem_custodian_file} {self.cores}"

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        atoms.charge = self.charge
        atoms.spin_multiplicity = self.spin_multiplicity
        mol = AseAtomsAdaptor.get_molecule(atoms)
        if self.prev_orbital_coeffs is not None:
            with open("53.0", mode="wb") as file:
                for val in self.prev_orbital_coeffs:
                    data = struct.pack("d", val)
                    file.write(data)
            if "overwrite_inputs" not in self.qchem_input_params:
                self.qchem_input_params["overwrite_inputs"] = {}
            if "rem" not in self.qchem_input_params["overwrite_inputs"]:
                self.qchem_input_params["overwrite_inputs"]["rem"] = {}
            if "scf_guess" not in self.qchem_input_params["overwrite_inputs"]["rem"]:
                self.qchem_input_params["overwrite_inputs"]["rem"]["scf_guess"] = "read"
        qcin = ForceSet(mol, qchem_version=6, **self.qchem_input_params)
        qcin.write("mol.qin")

    def read_results(self):
        data = QCOutput("mol.qout").data
        self.results["energy"] = data["final_energy"] * units.Hartree
        tmp_grad_data = []
        # Read the gradient scratch file in 8 byte chunks
        with zopen("131.0", mode="rb") as file:
            binary = file.read()
            tmp_grad_data.extend(
                struct.unpack("d", binary[ii * 8 : (ii + 1) * 8])[0]
                for ii in range(len(binary) // 8)
            )
        grad = [
            [
                float(tmp_grad_data[ii * 3]),
                float(tmp_grad_data[ii * 3 + 1]),
                float(tmp_grad_data[ii * 3 + 2]),
            ]
            for ii in range(len(tmp_grad_data) // 3)
        ]
        # Ensure that the scratch values match the correct values from the
        # output file but with higher precision
        if data["pcm_gradients"] is not None:
            gradient = data["pcm_gradients"][0]
        else:
            gradient = data["gradients"][0]
        for ii, subgrad in enumerate(grad):
            for jj, val in enumerate(subgrad):
                if abs(gradient[ii, jj] - val) > 1e-6:
                    raise ValueError(
                        "Difference between gradient value in scratch file vs. output file should not be this large."
                    )
                gradient[ii, jj] = val
        # Convert gradient to force + deal with units
        self.results["forces"] = gradient * (-units.Hartree / units.Bohr)
        self.prev_orbital_coeffs = []
        # Read orbital coefficients scratch file in 8 byte chunks
        with zopen("53.0", mode="rb") as file:
            binary = file.read()
            self.prev_orbital_coeffs.extend(
                struct.unpack("d", binary[ii * 8 : (ii + 1) * 8])[0]
                for ii in range(len(binary) // 8)
            )



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, TypeVar

    Job = TypeVar("Job")
    Flow = TypeVar("Flow")
    Subflow = TypeVar("Subflow")


def job(_func: Callable | None = None, **kwargs) -> Job:  # sourcery skip
    """
    Decorator for individual compute jobs. This is a @job decorator.

    job = ct.electron [Covalent] = python_app [Parsl] = job [Jobflow]
    = task [redun].

    The wrapped function gets a new kwarg, `decorator_kwargs`, that can be used
    to modify the workflow engine decorator keyword arguments. The wrapped
    function can also be accessed without the decorator by calling the
    `.__wrapped__` attribute.

    Parameters
    ----------
    _func
        The function to decorate. This is not meant to be supplied by the user.
    **kwargs
        Keyword arguments to pass to the workflow engine decorator.

    Returns
    -------
    Job
        The @job-decorated function.
    """

    @functools.wraps(_func)
    def _inner(*f_args, decorator_kwargs: dict | None = None, **f_kwargs) -> Any:
        """
        The @job-decorated function.

        Parameters
        ----------
        *f_args
            Positional arguments to the function, if any.
        decorator_kwargs
            Keyword arguments to pass to the workflow engine decorator.
        **f_kwargs
            Keyword arguments to the function, if any.

        Returns
        -------
        Any
            The output of the @job-decorated function.
        """

        from quacc import SETTINGS

        if decorator_kwargs is None:
            decorator_kwargs = kwargs

        wflow_engine = SETTINGS.WORKFLOW_ENGINE
        if wflow_engine == "covalent":
            import covalent as ct

            decorated = ct.electron(_func, **decorator_kwargs)
        elif wflow_engine == "jobflow":
            from jobflow import job as jf_job

            decorated = jf_job(_func, **decorator_kwargs)
        elif wflow_engine == "parsl":
            from parsl import python_app

            decorated = python_app(_func, **decorator_kwargs)
        elif wflow_engine == "redun":
            from redun import task

            decorated = task(_func, **decorator_kwargs)
        else:
            decorated = _func

        return decorated(*f_args, **f_kwargs)

    if _func is None:

        def decorator(_f):
            return job(_f, **kwargs)

        return decorator

    return _inner


def flow(_func: Callable | None = None, **kwargs) -> Flow:  # sourcery skip
    """
    Decorator for workflows, which consist of at least one compute job. This is
    a @flow decorator.

    flow = ct.lattice/ct.dispatch(ct.lattice) [Covalent] = task [redun].
    For Parsl and Jobflow, the decorator returns the original function, unchanged.

    The wrapped function gets a new kwarg, `decorator_kwargs`, that can be used
    to modify the workflow engine decorator keyword arguments. The wrapped
    function can also be accessed without the decorator by calling the
    `.__wrapped__` attribute.

    Parameters
    ----------
    _func
        The function to decorate. This is not meant to be supplied by the user.
    **kwargs
        Keyword arguments to pass to the decorator.

    Returns
    -------
    Flow
        The @flow-decorated function.
    """

    @functools.wraps(_func)
    def _inner(
        *f_args,
        decorator_kwargs: dict | None = None,
        dispatch_kwargs: dict | None = None,
        **f_kwargs,
    ) -> Any:
        """
        The @flow-decorated function.

        Parameters
        ----------
        *f_args
            Positional arguments to the function, if any.
        decorator_kwargs
            Keyword arguments to pass to the workflow engine decorator.
        dispatch_kwargs
            Keyword arguments to pass to ct.dispatch() if Covalent is used.
        **f_kwargs
            Keyword arguments to the function, if any.

        Returns
        -------
        Any
            The output of the @flow-decorated function.
        """
        from quacc import SETTINGS

        if decorator_kwargs is None:
            decorator_kwargs = kwargs

        wflow_engine = SETTINGS.WORKFLOW_ENGINE

        dispatch_kwargs = dispatch_kwargs or {}
        if dispatch_kwargs and wflow_engine != "covalent":
            raise ValueError("The `dispatch_kwargs` argument only works with Covalent.")

        if wflow_engine == "covalent":
            import covalent as ct

            try:
                return ct.dispatch(ct.lattice(_func, **decorator_kwargs))(
                    *f_args, **f_kwargs
                )
            except (AttributeError, TypeError):
                return ct.lattice(_func, **dispatch_kwargs, **decorator_kwargs)(
                    *f_args, **f_kwargs
                )
        elif wflow_engine == "redun":
            from redun import task

            decorated = task(_func, **decorator_kwargs)
        else:
            decorated = _func

        return decorated(*f_args, **f_kwargs)

    if _func is None:

        def decorator(_f):
            return flow(_f, **kwargs)

        return decorator

    return _inner


def subflow(_func: Callable | None = None, **kwargs) -> Subflow:  # sourcery skip
    """
    Decorator for (dynamic) sub-workflows. This is a @subflow decorator.

    subflow = ct.electron(ct.lattice) [Covalent] = join_app [Parsl] = task [Redun].
    For Jobflow, the decorator returns the original function.

    The wrapped function gets a new kwarg, `decorator_kwargs`, that can be used
    to modify the workflow engine decorator keyword arguments. The wrapped
    function can also be accessed without the decorator by calling the
    `.__wrapped__` attribute.

    Parameters
    ----------
    _func
        The function to decorate. This is not meant to be supplied by the user.
    **kwargs
        Keyword arguments to pass to the decorator.

    Returns
    -------
    callable
        The decorated function.
    """

    @functools.wraps(_func)
    def _inner(*f_args, decorator_kwargs: dict | None = None, **f_kwargs) -> Any:
        """
        The @subflow-decorated function.

        Parameters
        ----------
        *f_args
            Positional arguments to the function, if any.
        decorator_kwargs
            Keyword arguments to pass to the workflow engine decorator.
        **f_kwargs
            Keyword arguments to the function, if any.

        Returns
        -------
        Any
            The output of the @subflow-decorated function.
        """
        from quacc import SETTINGS

        if decorator_kwargs is None:
            decorator_kwargs = kwargs

        wflow_engine = SETTINGS.WORKFLOW_ENGINE
        if wflow_engine == "covalent":
            import covalent as ct

            decorated = ct.electron(ct.lattice(_func), **decorator_kwargs)
        elif wflow_engine == "parsl":
            from parsl import join_app

            decorated = join_app(_func, **decorator_kwargs)
        elif wflow_engine == "redun":
            from redun import task

            decorated = task(_func, **decorator_kwargs)
        else:
            decorated = _func

        return decorated(*f_args, **f_kwargs)

    if _func is None:

        def decorator(_f):
            return subflow(_f, **kwargs)

        return decorator

    return _inner
